# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for View."""

import json
import os
import pathlib
from typing import List, Tuple

from PyQt5.QtCore import QEvent, QPoint, QPointF, Qt
from PyQt5.QtGui import QKeyEvent, QMouseEvent, QPainter, QWheelEvent, QContextMenuEvent
from PyQt5.QtWidgets import QGraphicsView, QMenu
from PyQt5.sip import isdeleted


from pyflow.scene import Scene
from pyflow.core.socket import Socket
from pyflow.core.edge import Edge
from pyflow.blocks.block import Block
from pyflow.blocks.codeblock import CodeBlock
from pyflow.blocks import __file__ as BLOCK_INIT_PATH

BLOCK_PATH = pathlib.Path(BLOCK_INIT_PATH).parent
BLOCKFILES_PATH = os.path.join(BLOCK_PATH, "blockfiles")

EPS: float = 1e-10  # To check if blocks are of size 0


class View(QGraphicsView):

    """View for the Window."""

    MODE_NOOP = 0
    MODE_EDGE_DRAG = 1
    MODE_EDITING = 2

    MODES = {
        "NOOP": MODE_NOOP,
        "EDGE_DRAG": MODE_EDGE_DRAG,
        "EDITING": MODE_EDITING,
    }

    def __init__(
        self,
        scene: Scene,
        parent=None,
        zoom_step: float = 1.25,
        zoom_min: float = 0.2,
        zoom_max: float = 5,
    ):
        super().__init__(parent=parent)
        self.mode = self.MODE_NOOP
        self.zoom = 1
        self.zoom_step, self.zoom_min, self.zoom_max = zoom_step, zoom_min, zoom_max

        self.edge_drag = None
        self.lastMousePos = QPointF(0, 0)
        self.currentSelectedBlock = None

        self.init_ui()
        self.setScene(scene)

    def init_ui(self):
        """Initialize the custom  View UI."""
        # Antialiasing
        self.setRenderHints(
            QPainter.RenderHint.Antialiasing
            | QPainter.RenderHint.HighQualityAntialiasing
            | QPainter.RenderHint.TextAntialiasing
            | QPainter.RenderHint.SmoothPixmapTransform
        )
        # Better Update
        self.setViewportUpdateMode(QGraphicsView.ViewportUpdateMode.FullViewportUpdate)
        # Remove scroll bars
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        # Zoom on cursor
        self.setTransformationAnchor(QGraphicsView.ViewportAnchor.AnchorUnderMouse)
        # Selection box
        self.setDragMode(QGraphicsView.DragMode.RubberBandDrag)

    def scene(self) -> Scene:
        """Get current Scene."""
        return super().scene()

    def mousePressEvent(self, event: QMouseEvent):
        """Dispatch Qt's mousePress events to corresponding functions below."""
        if event.button() == Qt.MouseButton.MiddleButton:
            self.middleMouseButtonPress(event)
        elif event.button() == Qt.MouseButton.LeftButton:
            self.leftMouseButtonPress(event)
        else:
            super().mousePressEvent(event)

    def mouseReleaseEvent(self, event: QMouseEvent):
        """Dispatch Qt's mouseRelease events to corresponding functions below."""
        if event.button() == Qt.MouseButton.MiddleButton:
            self.middleMouseButtonRelease(event)
        elif event.button() == Qt.MouseButton.LeftButton:
            self.leftMouseButtonRelease(event)
        else:
            super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event: QMouseEvent) -> None:
        """View reaction to mouseMoveEvent."""
        self.lastMousePos = self.mapToScene(event.pos())
        self.drag_edge(event, "move")
        if event is not None:
            super().mouseMoveEvent(event)

    def leftMouseButtonPress(self, event: QMouseEvent):
        """View reaction to leftMouseButtonPress event."""
        # If clicked on a block, bring it forward.
        item_at_click = self.itemAt(event.pos())
        if item_at_click is not None:
            while item_at_click.parentItem() is not None:
                if isinstance(item_at_click, Block):
                    break
                item_at_click = item_at_click.parentItem()

            if isinstance(item_at_click, Block):
                self.bring_block_forward(item_at_click)

        # If clicked on a socket, start dragging an edge.
        event = self.drag_edge(event, "press")
        if event is not None:
            super().mousePressEvent(event)

    def leftMouseButtonRelease(self, event: QMouseEvent):
        """View reaction to leftMouseButtonRelease event."""
        event = self.drag_edge(event, "release")
        if event is not None:
            super().mouseReleaseEvent(event)

    def middleMouseButtonPress(self, event: QMouseEvent):
        """View reaction to middleMouseButtonPress event."""
        if self.itemAt(event.pos()) is None:
            event = self.drag_scene(event, "press")
        super().mousePressEvent(event)

    def middleMouseButtonRelease(self, event: QMouseEvent):
        """View reaction to middleMouseButtonRelease event."""
        event = self.drag_scene(event, "release")
        super().mouseReleaseEvent(event)
        self.setDragMode(QGraphicsView.DragMode.RubberBandDrag)

    def centerView(self, x: float, y: float):
        """Move the view so that the position (x,y) is centered."""
        hsb = self.horizontalScrollBar()
        vsb = self.verticalScrollBar()
        hsb.setValue(x * self.zoom - self.width() / 2)
        vsb.setValue(y * self.zoom - self.height() / 2)

    def moveToItems(self) -> bool:
        """Ajust zoom and position to make selected items visible.

        If no item is selected, make the whole graph visible instead.

        Returns:
            True if the event was handled, False otherwise.
        """

        # The focusItem has priority for this event
        if self.scene().focusItem() is not None:
            return False

        items = self.scene().items()

        # If items are selected, overwride the behvaior
        if len(self.scene().selectedItems()) > 0:
            items = self.scene().selectedItems()

        code_blocks: List[Block] = [i for i in items if isinstance(i, Block)]

        if not code_blocks:
            return False

        # Get the blocks with min and max x and y coordinates

        min_x: float = min(block.x() for block in code_blocks)
        min_y: float = min(block.y() for block in code_blocks)
        max_x: float = max(block.x() + block.width for block in code_blocks)
        max_y: float = max(block.y() + block.height for block in code_blocks)

        center_x: float = (min_x + max_x) / 2
        center_y: float = (min_y + max_y) / 2

        # Determines the required zoom level

        if max_x - min_x < EPS or max_y - min_y < EPS:
            # Handle the case where there is no block
            return False

        required_zoom_x: float = self.width() / (max_x - min_x)
        required_zoom_y: float = self.height() / (max_y - min_y)

        # Operate the zoom and the translation
        self.setZoom(min(required_zoom_x, required_zoom_y))
        self.centerView(center_x, center_y)
        return True

    def getDistanceToCenter(self, x: float, y: float) -> Tuple[float]:
        """Return the vector from the (x,y) position given to the center of the view."""
        ypos = self.verticalScrollBar().value()
        xpos = self.horizontalScrollBar().value()
        return (
            xpos - x * self.zoom + self.width() / 2,
            ypos - y * self.zoom + self.height() / 2,
        )

    def moveViewOnArrow(self, event: QKeyEvent) -> bool:
        """
        View reaction to an arrow key being pressed.
        Returns True if the event was handled.
        """
        # The focusItem has priority for this event if it is a source editor
        if self.scene().focusItem() is not None:
            parent = self.scene().focusItem().parentItem()
            if isinstance(parent, CodeBlock) and parent.source_editor.hasFocus():
                return False

        n_selected_items = len(self.scene().selectedItems())
        if n_selected_items > 1:
            return False

        code_blocks = [
            i
            for i in self.scene().items()
            if isinstance(i, Block) and not i.isSelected()
        ]

        reference = None
        if n_selected_items == 1 and isinstance(self.scene().selectedItems()[0], Block):
            selected_item = self.scene().selectedItems()[0]
            reference = QPoint(
                selected_item.x() + selected_item.width / 2,
                selected_item.y() + selected_item.height / 2,
            )
            self.scene().clearSelection()

        dist_array = []
        for block in code_blocks:
            block_center_x = block.x() + block.width / 2
            block_center_y = block.y() + block.height / 2
            if reference is None:
                xdist, ydist = self.getDistanceToCenter(block_center_x, block_center_y)
            else:
                xdist = reference.x() - block_center_x
                ydist = reference.y() - block_center_y
            dist_array.append((block_center_x, block_center_y, -xdist, -ydist))

        def in_region(x, y, key):
            up_right = x / self.width() - y / self.height() >= 0
            down_right = x / self.width() + y / self.height() >= 0
            if key == Qt.Key.Key_Up:
                return up_right and not down_right
            if key == Qt.Key.Key_Down:
                return not up_right and down_right
            if key == Qt.Key.Key_Left:
                return not up_right and not down_right
            if key == Qt.Key.Key_Right:
                return up_right and down_right

        key_id = event.key()
        dist_array = filter(lambda pos: in_region(pos[2], pos[3], key_id), dist_array)
        dist_array = list(dist_array)
        if not dist_array:
            return False

        def oriented_distance(x, y, key):
            if key in (Qt.Key.Key_Down, Qt.Key.Key_Up):
                return abs(y) / self.height() + (x / self.width()) ** 2
            if key in (Qt.Key.Key_Left, Qt.Key.Key_Right):
                return abs(x) / self.width() + (y / self.height()) ** 2

        dist_array.sort(key=lambda pos: oriented_distance(pos[2], pos[3], key_id))
        block_center_x, block_center_y, _, _ = dist_array[0]

        item_to_navigate = self.scene().itemAt(
            block_center_x, block_center_y, self.transform()
        )
        if isinstance(item_to_navigate.parentItem(), Block):
            item_to_navigate.parentItem().setSelected(True)

        self.centerView(block_center_x, block_center_y)
        return True

    def keyPressEvent(self, event: QKeyEvent):
        """View reaction to a key being pressed."""
        key_id = event.key()
        if key_id in [
            Qt.Key.Key_Up,
            Qt.Key.Key_Down,
            Qt.Key.Key_Left,
            Qt.Key.Key_Right,
        ]:
            if self.moveViewOnArrow(event):
                return

        super().keyPressEvent(event)

    def retreiveBlockTypes(self) -> List[Tuple[str]]:
        """Retreive the list of stored blocks."""
        block_type_files = os.listdir(BLOCKFILES_PATH)
        block_types = []
        for blockfile_name in block_type_files:
            filepath = os.path.join(BLOCKFILES_PATH, blockfile_name)
            with open(filepath, "r", encoding="utf-8") as file:
                data = json.loads(file.read())
                title = "New Block"
                if "title" in data:
                    title = f"New {data['title']} Block"
                    if data["title"] == "Empty":
                        block_types[:0] = [(filepath, title)]
                    else:
                        block_types.append((filepath, title))
        return block_types

    def contextMenuEvent(self, event: QContextMenuEvent):
        """Displays the context menu when inside a view."""
        super().contextMenuEvent(event)
        # If somebody has already accepted the event, don't handle it.
        if event.isAccepted():
            return
        event.setAccepted(True)

        menu = QMenu(self)
        actionPool = []
        for filepath, block_name in self.retreiveBlockTypes():
            actionPool.append((filepath, menu.addAction(block_name)))

        selectedAction = menu.exec_(self.mapToGlobal(event.pos()))
        for filepath, action in actionPool:
            if action == selectedAction:
                p = self.mapToScene(event.pos())
                self.scene().create_block_from_file(filepath, p.x(), p.y())

    def wheelEvent(self, event: QWheelEvent):
        """Handles zooming with mouse wheel events."""
        if Qt.Modifier.CTRL == int(event.modifiers()):
            # calculate zoom
            if event.angleDelta().y() > 0:
                zoom_factor = self.zoom_step
            else:
                zoom_factor = 1 / self.zoom_step

            new_zoom = self.zoom * zoom_factor
            if self.zoom_min < new_zoom < self.zoom_max:
                self.setZoom(new_zoom)
        else:
            super().wheelEvent(event)

    def setZoom(self, new_zoom: float):
        """Set the zoom to the appropriate level."""
        zoom_factor = new_zoom / self.zoom
        self.scale(zoom_factor, zoom_factor)
        self.zoom = new_zoom

    def deleteSelected(self):
        """Delete selected items from the current scene."""
        scene = self.scene()
        for selected_item in scene.selectedItems():
            selected_item.remove()
        scene.history.checkpoint("Delete selected elements", set_modified=True)

    def bring_block_forward(self, block: Block):
        """Move the selected block in front of other blocks.

        Args:
            block: Block to bring forward.

        """
        if self.currentSelectedBlock is not None and not isdeleted(
            self.currentSelectedBlock
        ):
            self.currentSelectedBlock.setZValue(0)
        block.setZValue(1)
        self.currentSelectedBlock = block

    def drag_scene(self, event: QMouseEvent, action="press"):
        """Drag the scene around."""
        if action == "press":
            releaseEvent = QMouseEvent(
                QEvent.Type.MouseButtonRelease,
                event.localPos(),
                event.screenPos(),
                Qt.MouseButton.LeftButton,
                Qt.MouseButton.NoButton,
                event.modifiers(),
            )
            super().mouseReleaseEvent(releaseEvent)
            self.setDragMode(QGraphicsView.DragMode.ScrollHandDrag)
            return QMouseEvent(
                event.type(),
                event.localPos(),
                event.screenPos(),
                Qt.MouseButton.LeftButton,
                event.buttons() | Qt.MouseButton.LeftButton,
                event.modifiers(),
            )
        return QMouseEvent(
            event.type(),
            event.localPos(),
            event.screenPos(),
            Qt.MouseButton.LeftButton,
            event.buttons() & ~Qt.MouseButton.LeftButton,
            event.modifiers(),
        )

    def drag_edge(self, event: QMouseEvent, action="press"):
        """Create an edge by drag and drop."""
        item_at_click = self.itemAt(event.pos())
        scene = self.scene()
        if action == "press":
            if (
                isinstance(item_at_click, Socket)
                and self.mode != self.MODE_EDGE_DRAG
                and item_at_click.socket_type != "input"
            ):
                self.mode = self.MODE_EDGE_DRAG
                self.edge_drag = Edge(
                    source_socket=item_at_click,
                    destination=self.mapToScene(event.pos()),
                )
                scene.addItem(self.edge_drag)
                return
        elif action == "release":
            if self.mode == self.MODE_EDGE_DRAG:
                if (
                    isinstance(item_at_click, Socket)
                    and item_at_click is not self.edge_drag.source_socket
                    and item_at_click.socket_type != "output"
                ):
                    self.edge_drag.destination_socket = item_at_click
                    scene.history.checkpoint(
                        "Created edge by dragging", set_modified=True
                    )
                else:
                    self.edge_drag.remove()
                self.edge_drag = None
                self.mode = self.MODE_NOOP
        elif action == "move":
            if self.mode == self.MODE_EDGE_DRAG:
                self.edge_drag.destination = self.mapToScene(event.pos())
        return event

    def set_mode(self, mode: str):
        """Change the view mode.

        Args:
            mode: Mode key to change to, must in present in MODES.

        """
        self.mode = self.MODES[mode]

    def is_mode(self, mode: str):
        """Return True if the view is in the given mode.

        Args:
            mode: Mode key to compare to, must in present in MODES.

        """
        return self.mode == self.MODES[mode]



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the abstract ExecutableBlock class.

An abstract block that allows for execution, like CodeBlocks and Sliders.

"""

from typing import OrderedDict
from abc import abstractmethod
from PyQt5.QtCore import QTimer
from PyQt5.QtWidgets import QApplication

from pyflow.blocks.block import Block
from pyflow.core.socket import Socket


class ExecutableBlock(Block):

    """
    Executable Block

    This block type is not meant to be instanciated !

    It's an abstract class that represents blocks that can be executed like:
    - CodeBlock
    - Slider

    """

    def __init__(self, **kwargs):
        """
        Create a new executable block.
        Do not call this method except when inheriting from this class.
        """
        super().__init__(**kwargs)

        self.has_been_run = False
        self._run_state = 0

        # Each element is a list of blocks/edges to be animated
        # Running will paint each element one after the other
        self.transmitting_queue = []
        # Controls the duration of the visual flow animation
        self.transmitting_duration = 500

        # Add execution flow sockets
        exe_sockets = (
            Socket(self, socket_type="input", flow_type="exe"),
            Socket(self, socket_type="output", flow_type="exe"),
        )
        for socket in exe_sockets:
            self.add_socket(socket)

        if type(self) == ExecutableBlock:
            raise RuntimeError("ExecutableBlock should not be instanciated directly")

    def has_input(self) -> bool:
        """Checks whether a block has connected input blocks."""
        for input_socket in self.sockets_in:
            if input_socket.edges:
                return True
        return False

    def has_output(self) -> bool:
        """Checks whether a block has connected output blocks."""
        for output_socket in self.sockets_out:
            if output_socket.edges:
                return True
        return False

    def run_code(self):
        """Run the code in the block."""

        # Queue the code to execute
        code = self.source
        kernel = self.scene().kernel
        kernel.execution_queue.append((self, code))

        if kernel.busy is False:
            kernel.run_queue()
        self.has_been_run = True

    def execution_finished(self):
        """Reset the text of the run buttons."""
        self.run_state = 0
        self.blocks_to_run = []

    def _interrupt_execution(self):
        """Interrupt an execution, reset the blocks in the queue."""
        for block, _ in self.scene().kernel.execution_queue:
            # Reset the blocks that have not been run
            block.reset_has_been_run()
            block.execution_finished()
        # Clear kernel execution queue
        self.scene().kernel.execution_queue = []
        # Interrupt the kernel
        self.scene().kernel.kernel_manager.interrupt_kernel()
        # Clear local execution queue
        self.blocks_to_run = []

    def transmitting_animation_in(self):
        """
        Animate the visual flow
        Set color to transmitting and set a timer before switching to normal
        """
        for elem in self.transmitting_queue[0]:
            # Set color to transmitting
            elem.run_state = 2
        QApplication.processEvents()
        QTimer.singleShot(self.transmitting_delay, self.transmitting_animation_out)

    def transmitting_animation_out(self):
        """
        Animate the visual flow
        After the timer, set color to normal and move on with the queue
        """
        for elem in self.transmitting_queue[0]:
            # Reset color only if the block will not be run
            if hasattr(elem, "has_been_run"):
                if elem.has_been_run is True:
                    elem.run_state = 0
            else:
                elem.run_state = 0

        QApplication.processEvents()
        self.transmitting_queue.pop(0)
        if self.transmitting_queue:
            # If the queue is not empty, move forward in the animation
            self.transmitting_animation_in()
        else:
            # Else, run the blocks in the self.blocks_to_run
            self.run_blocks()

    def custom_bfs(self, start_node, reverse=False):
        """
        Graph traversal in BFS to find the blocks that are connected to the start_node

        Args:
            start_node (Block): The block to start the traversal from
            reverse (bool): If True, traverse in the direction of outputs

        Returns:
            list: Blocks to run in topological order (reversed)
            list: each element is a list of blocks/edges to animate in order
        """
        # Blocks to run in topological order
        blocks_to_run = []
        # List of lists of blocks/edges to animate in order
        to_transmit = [[start_node]]

        to_visit = [start_node]
        while to_visit:
            # Remove duplicates
            to_visit = list(set(to_visit))

            # Gather connected edges
            edges_to_visit = []
            for block in to_visit:
                blocks_to_run.append(block)
                if not reverse:
                    for input_socket in block.sockets_in:
                        for edge in input_socket.edges:
                            edges_to_visit.append(edge)
                else:
                    for output_socket in block.sockets_out:
                        for edge in output_socket.edges:
                            edges_to_visit.append(edge)
            to_transmit.append(edges_to_visit)

            # Gather connected blocks
            to_visit = []
            for edge in edges_to_visit:
                if not reverse:
                    to_visit.append(edge.source_socket.block)
                else:
                    to_visit.append(edge.destination_socket.block)
            to_transmit.append(to_visit)

        # Remove start node
        blocks_to_run.pop(0)

        return blocks_to_run, to_transmit

    def right_traversal(self):
        """
        Custom graph traversal utility
        Returns blocks/edges that will potentially be run by run_right
        from closest to farthest from self

        Returns:
            list: each element is a list of blocks/edges to animate in order
        """
        # Result
        to_transmit = [[self]]

        # To check if a block has been visited
        visited = []
        # We need to visit the inputs of these blocks
        to_visit_input = [self]
        # We need to visit the outputs of these blocks
        to_visit_output = [self]

        # Next stage to put in to_transmit
        next_edges = []
        next_blocks = []

        while to_visit_input or to_visit_output:
            for block in to_visit_input.copy():
                # Check input edges and blocks
                for input_socket in block.sockets_in:
                    for edge in input_socket.edges:
                        if edge not in visited:
                            next_edges.append(edge)
                        visited.append(edge)
                        input_block = edge.source_socket.block
                        to_visit_input.append(input_block)
                        if input_block not in visited:
                            next_blocks.append(input_block)
                        visited.append(input_block)
                to_visit_input.remove(block)
            for block in to_visit_output.copy():
                # Check output edges and blocks
                for output_socket in block.sockets_out:
                    for edge in output_socket.edges:
                        if edge not in visited:
                            next_edges.append(edge)
                        visited.append(edge)
                        output_block = edge.destination_socket.block
                        to_visit_input.append(output_block)
                        to_visit_output.append(output_block)
                        if output_block not in visited:
                            next_blocks.append(output_block)
                        visited.append(output_block)
                to_visit_output.remove(block)

            # Add the next stage to to_transmit
            to_transmit.append(next_edges)
            to_transmit.append(next_blocks)

            # Reset next stage
            next_edges = []
            next_blocks = []

        return to_transmit

    def run_blocks(self):
        """Run a list of blocks."""
        for block in self.blocks_to_run[::-1]:
            if not block.has_been_run:
                block.run_code()
        if not self.has_been_run:
            self.run_code()

    def run_left(self):
        """Run all of the block's dependencies and then run the block."""

        # Reset has_been_run to make sure that the self is run again
        self.has_been_run = False

        # To avoid crashing when spamming the button
        if self.transmitting_queue:
            return

        # Gather dependencies
        blocks_to_run, to_transmit = self.custom_bfs(self)
        self.blocks_to_run = blocks_to_run

        # Set the transmitting queue
        self.transmitting_queue = to_transmit
        # Set delay so that the transmitting animation has fixed total duration
        self.transmitting_delay = int(
            self.transmitting_duration / len(self.transmitting_queue)
        )
        # Start transmitting animation
        self.transmitting_animation_in()

    def run_right(self):
        """Run all of the output blocks and all their dependencies."""

        # To avoid crashing when spamming the button
        if self.transmitting_queue:
            return

        # Create transmitting queue
        self.transmitting_queue = self.right_traversal()

        # Gather outputs
        blocks_to_run, _ = self.custom_bfs(self, reverse=True)
        self.blocks_to_run = [self] + blocks_to_run

        # For each output found
        for block in self.blocks_to_run.copy()[::-1]:
            # Gather dependencies
            new_blocks_to_run, _ = self.custom_bfs(block)
            self.blocks_to_run += new_blocks_to_run

        # Set delay so that the transmitting animation has fixed total duration
        self.transmitting_delay = int(
            self.transmitting_duration / len(self.transmitting_queue)
        )
        # Start transmitting animation
        self.transmitting_animation_in()

    def reset_has_been_run(self):
        """Called when the output is an error."""
        self.has_been_run = False

    @property
    @abstractmethod
    def source(self) -> str:
        """Source code."""
        raise NotImplementedError("source(self) should be overriden")

    @source.setter
    @abstractmethod
    def source(self, value: str):
        raise NotImplementedError("source(self) should be overriden")

    @property
    def run_state(self) -> int:
        """Run state.

        Describe the current state of the ExecutableBlock:
            - 0: idle.
            - 1: running.
            - 2: transmitting.

        """
        return self._run_state

    @run_state.setter
    def run_state(self, value: int):
        self._run_state = value
        # Update to force repaint
        self.update()

    def handle_stdout(self, value: str):
        """Handle the stdout signal."""

    def handle_image(self, image: str):
        """Handle the image signal."""

    def serialize(self):
        """Return a serialized version of this block."""
        return super().serialize()

    def deserialize(
        self, data: OrderedDict, hashmap: dict = None, restore_id: bool = True
    ):
        """Restore a codeblock from it's serialized state."""
        super().deserialize(data, hashmap, restore_id)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the base Edge."""

from __future__ import annotations

from typing import Optional, OrderedDict

from PyQt5.QtCore import QPointF, Qt
from PyQt5.QtGui import QColor, QPainter, QPainterPath, QPen
from PyQt5.QtWidgets import (
    QGraphicsPathItem,
    QStyleOptionGraphicsItem,
    QWidget,
)

from pyflow.core.serializable import Serializable
from pyflow.core.socket import Socket


class Edge(QGraphicsPathItem, Serializable):

    """Base class for directed edges in Pyflow."""

    DEFAULT_DATA = {"path_type": "bezier"}
    MANDATORY_FIELDS = {"source", "destination"}

    def __init__(
        self,
        edge_width: float = 4.0,
        path_type=DEFAULT_DATA["path_type"],
        edge_color="#001000",
        edge_selected_color="#00ff00",
        edge_running_color="#FF0000",
        edge_transmitting_color="#00ff00",
        source: QPointF = QPointF(0, 0),
        destination: QPointF = QPointF(0, 0),
        source_socket: Socket = None,
        destination_socket: Socket = None,
    ):
        """Base class for edges in Pyflow.

        Args:
            edge_width: Width of the edge.
            path_type: Type of path, one of ('direct', 'bezier').
            edge_color: Color of the edge.
            edge_selected_color: Color of the edge when it is selected.
            source: Source point of the directed edge.
            destination: Destination point of the directed edge.
            source_socket: Source socket of the directed edge, overrides source.
            destination_socket: Destination socket of the directed edge, overrides destination.

        """

        Serializable.__init__(self)
        QGraphicsPathItem.__init__(self, parent=None)
        self._pen = QPen(QColor(edge_color))
        self._pen.setWidthF(edge_width)

        self._pen_dragging = QPen(QColor(edge_color))
        self._pen_dragging.setWidthF(edge_width)
        self._pen_dragging.setStyle(Qt.PenStyle.DashLine)

        self._pen_selected = QPen(QColor(edge_selected_color))
        self._pen_selected.setWidthF(edge_width)

        self._pen_running = QPen(QColor(edge_running_color))
        self._pen_running.setWidthF(edge_width)

        self._pen_transmitting = QPen(QColor(edge_transmitting_color))
        self._pen_transmitting.setWidthF(edge_width)

        self.pens = [self._pen, self._pen_running, self._pen_transmitting]

        # 0 for normal, 1 for running, 2 for transmitting
        self.run_state = 0

        self.setFlag(QGraphicsPathItem.GraphicsItemFlag.ItemIsSelectable)
        self.setZValue(-1)

        self.path_type = path_type

        self.source_socket = source_socket
        self.destination_socket = destination_socket

        self._source = source
        self._destination = destination
        self.update_path()

    def remove_from_socket(self, socket_type="source"):
        """Remove the edge from the sockets it is snaped to on the given socket_type.

        Args:
            socket_type: One of ('source', 'destination').

        """
        socket_name = f"{socket_type}_socket"
        socket = getattr(self, socket_name, Socket)
        if socket is not None:
            socket.remove_edge(self)
            setattr(self, socket_name, None)

    def remove_from_sockets(self):
        """Remove the edge from all sockets it is snaped to."""
        self.remove_from_socket("source")
        self.remove_from_socket("destination")

    def remove(self):
        """Remove the edge from the scene in which it is drawn."""
        scene = self.scene()
        if scene is not None:
            self.remove_from_sockets()
            scene.removeItem(self)

    def paint(
        self,
        painter: QPainter,
        option: QStyleOptionGraphicsItem,  # pylint:disable=unused-argument
        widget: Optional[QWidget] = None,
    ):  # pylint:disable=unused-argument
        """Paint the edge."""
        self.update_path()
        if self.isSelected():
            pen = self._pen_selected
        elif self.destination_socket is None:
            pen = self._pen_dragging
        else:
            pen = self.pens[self.run_state]
        painter.setPen(pen)
        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.drawPath(self.path())

    def update_path(self):
        """Update the edge path depending on the path_type."""
        path = QPainterPath(self.source)
        if self.path_type == "direct":
            path.lineTo(self.destination)
        elif self.path_type == "bezier":
            sx, sy = self.source.x(), self.source.y()
            dx, dy = self.destination.x(), self.destination.y()
            mid_dist = (dx - sx) / 2
            path.cubicTo(sx + mid_dist, sy, dx - mid_dist, dy, dx, dy)
        else:
            raise NotImplementedError(f"Unknowed path type: {self.path_type}")
        self.setPath(path)

    @property
    def source(self) -> QPointF:
        """Source point of the directed edge."""
        if self.source_socket is not None:
            return self.source_socket.scenePos()
        return self._source

    @source.setter
    def source(self, value: QPointF):
        self._source = value
        try:
            self.update_path()
        except AttributeError:
            pass

    @property
    def source_socket(self) -> Socket:
        """Source socket of the directed edge."""
        return self._source_socket

    @source_socket.setter
    def source_socket(self, value: Socket):
        self._source_socket = value
        if value is not None:
            self.source_socket.add_edge(self, is_destination=False)
            self.source = value.scenePos()

    @property
    def destination(self) -> QPointF:
        """Destination point of the directed edge."""
        if self.destination_socket is not None:
            return self.destination_socket.scenePos()
        return self._destination

    @destination.setter
    def destination(self, value: QPointF):
        self._destination = value
        try:
            self.update_path()
        except AttributeError:
            pass

    @property
    def destination_socket(self) -> Socket:
        """Destination socket of the directed edge."""
        return self._destination_socket

    @destination_socket.setter
    def destination_socket(self, value: Socket):
        self._destination_socket = value
        if value is not None:
            self.destination_socket.add_edge(self, is_destination=True)
            self.destination = value.scenePos()

    def serialize(self) -> OrderedDict:
        return OrderedDict(
            [
                ("id", self.id),
                ("path_type", self.path_type),
                (
                    "source",
                    OrderedDict(
                        [
                            (
                                "block",
                                self.source_socket.block.id
                                if self.source_socket
                                else None,
                            ),
                            (
                                "socket",
                                self.source_socket.id if self.source_socket else None,
                            ),
                        ]
                    ),
                ),
                (
                    "destination",
                    OrderedDict(
                        [
                            (
                                "block",
                                self.destination_socket.block.id
                                if self.destination_socket
                                else None,
                            ),
                            (
                                "socket",
                                self.destination_socket.id
                                if self.destination_socket
                                else None,
                            ),
                        ]
                    ),
                ),
            ]
        )

    def deserialize(self, data: OrderedDict, hashmap: dict = None, restore_id=True):
        if restore_id and "id" in data:
            self.id = data["id"]

        self.complete_with_default(data)

        self.path_type = data["path_type"]
        try:
            self.source_socket = hashmap[data["source"]["socket"]]

            self.destination_socket = hashmap[data["destination"]["socket"]]
            self.update_path()
        except KeyError:
            self.remove()

    @property
    def run_state(self) -> int:
        """Run state.

        Describe the current state of the Edge:
            - 0: idle.
            - 1: running.
            - 2: transmitting.

        """
        return self._run_state

    @run_state.setter
    def run_state(self, value: int):
        self._run_state = value
        # Update to force repaint
        self.update()



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the  Blocks Widgets. """

from pyflow.blocks.widgets.blocksplitter import Splitter
from pyflow.blocks.widgets.blocktitle import Title
from pyflow.blocks.widgets.blocksizegrip import SizeGrip



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>
# pylint:disable=unused-argument

""" Module for the Drawing Block.

A block in which you can draw.

"""

from math import floor
import json
from typing import OrderedDict

from PyQt5.QtCore import Qt, pyqtSignal
from PyQt5.QtGui import QColor, QMouseEvent, QPaintEvent, QPainter
from PyQt5.QtWidgets import QPushButton, QWidget
from pyflow.blocks.executableblock import ExecutableBlock


eps = 1


class DrawableWidget(QWidget):
    """A drawable widget is a canvas like widget on which you can doodle."""

    on_value_changed = pyqtSignal()

    def __init__(self, parent: QWidget):
        """Create a new Drawable widget."""
        super().__init__(parent)
        self.setAttribute(Qt.WA_PaintOnScreen)
        self.pixel_width = 24
        self.pixel_height = 24
        self.color_buffer = []
        self.mouse_down = False
        for _ in range(self.pixel_width):
            self.color_buffer.append([])
            for _ in range(self.pixel_height):
                # 0 = white, 1 = black
                self.color_buffer[-1].append(0)

    def clearDrawing(self):
        """Clear the drawing."""
        for i in range(self.pixel_width):
            for j in range(self.pixel_height):
                self.color_buffer[i][j] = 0

    def paintEvent(self, evt: QPaintEvent):
        """Draw the content of the widget."""
        painter = QPainter(self)

        for i in range(self.pixel_width):
            self.color_buffer.append([])
            for j in range(self.pixel_height):
                w = self.width() / self.pixel_width
                h = self.height() / self.pixel_height
                painter.fillRect(
                    w * i,
                    h * j,
                    w + eps,
                    h + eps,
                    # hex color encoded as AARRGGBB
                    QColor.fromRgb(
                        0xFF000000 if self.color_buffer[i][j] else 0xFFFFFFFF
                    ),
                )

    def mouseMoveEvent(self, evt: QMouseEvent):
        """Change the drawing when dragging the mouse around."""
        if self.mouse_down:
            x = floor(evt.x() / self.width() * self.pixel_width)
            y = floor(evt.y() / self.height() * self.pixel_height)
            if 0 <= x < self.pixel_width and 0 <= y < self.pixel_height:
                self.color_buffer[x][y] = 1
                self.repaint()
                self.on_value_changed.emit()

    def mousePressEvent(self, evt: QMouseEvent):
        """Signal that the drawing starts."""
        self.mouse_down = True

    def mouseReleaseEvent(self, evt: QMouseEvent):
        """Signal that the drawing stops."""
        self.mouse_down = False


class DrawingBlock(ExecutableBlock):

    """An Block on which you can draw, to test your CNNs for example."""

    def __init__(self, **kwargs):
        """Create a new Block."""
        super().__init__(**kwargs)

        self.draw_area = DrawableWidget(self.root)
        self.draw_area.on_value_changed.connect(self.valueChanged)
        self.var_name = "drawing"

        self.splitter.addWidget(self.draw_area)  # QGraphicsView
        self.run_button = QPushButton("Clear", self.root)
        self.run_button.move(
            int(self.edge_size * 2),
            int(self.title_widget.height() + self.edge_size * 2),
        )
        self.run_button.setFixedSize(int(8 * self.edge_size), int(3 * self.edge_size))
        self.run_button.clicked.connect(self.draw_area.clearDrawing)
        self.holder.setWidget(self.root)

    @property
    def drawing(self):
        """A json-encoded representation of the drawing."""
        return json.dumps(self.draw_area.color_buffer)

    @drawing.setter
    def drawing(self, value: str):
        self.draw_area.color_buffer = json.loads(value)

    def serialize(self):
        """Return a serialized version of this widget."""
        base_dict = super().serialize()
        base_dict["drawing"] = self.drawing

        return base_dict

    def valueChanged(self):
        """Called when the content of the drawing block changes."""
        # Make sure that the slider is initialized before trying to run it.
        if self.scene() is not None:
            self.run_right()

    @property
    def source(self):
        """The "source code" of the drawingblock i.e an assignement to the drawing buffer."""
        python_code = f"{self.var_name} = {repr(self.draw_area.color_buffer)}"
        return python_code

    @source.setter
    def source(self, value: str):
        raise RuntimeError("The source of a drawingblock is read-only.")

    def deserialize(
        self, data: OrderedDict, hashmap: dict = None, restore_id: bool = True
    ):
        """Restore a markdown block from it's serialized state."""
        for dataname in ["drawing"]:
            if dataname in data:
                setattr(self, dataname, data[dataname])

        super().deserialize(data, hashmap, restore_id)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the ContainerBlock.

A block that can contain other blocks.

"""

from PyQt5.QtWidgets import QVBoxLayout
from pyflow.blocks.block import Block


class ContainerBlock(Block):
    """
    A block that can contain other blocks.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Defer import to prevent circular dependency.
        # Due to the overall structure of the code, this cannot be removed, as the
        # scene should be able to serialize blocks.
        # This is not due to bad code design and should not be removed.
        # pylint: disable=import-outside-toplevel, cyclic-import
        from pyflow.graphics.view import View
        from pyflow.scene.scene import Scene

        self.layout = QVBoxLayout(self.root)
        self.layout.setContentsMargins(
            self.edge_size * 2,
            self.title_widget.height() + self.edge_size * 2,
            self.edge_size * 2,
            self.edge_size * 2,
        )

        self.child_scene = Scene()
        self.child_view = View(self.child_scene)
        self.layout.addWidget(self.child_view)

        self.holder.setWidget(self.root)



# -*- coding: utf-8 -*-
# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Pyflow: An open-source tool for modular visual programing in python """

__appname__ = "Pyflow"
__author__ = "Mathïs Fédérico"
__version__ = "1.0.0-beta"



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the base PyFlow Widget."""

import os

from PyQt5.QtWidgets import QVBoxLayout, QWidget
from PyQt5.QtCore import Qt

from pyflow.scene import Scene
from pyflow.graphics.view import View


class Widget(QWidget):

    """Window for a graph visualisation."""

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setAttribute(Qt.WidgetAttribute.WA_DeleteOnClose)

        self.layout = QVBoxLayout()
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self.layout)

        # Graphics Scene
        self.scene = Scene()
        self.scene.addHasBeenModifiedListener(self.updateTitle)

        # Graphics View
        self.view = View(self.scene)
        self.layout.addWidget(self.view)

        self.savepath = None

    def updateTitle(self):
        """Update the window title."""
        if self.savepath is None:
            title = "New Graph"
        else:
            title = os.path.basename(self.savepath)
        if self.isModified():
            title += "*"
        self.setWindowTitle(title)

    def isModified(self) -> bool:
        """Return True if the scene has been modified, False otherwise."""
        return self.scene.has_been_modified

    @property
    def savepath(self):
        """Current cached file save path. Update window title when set."""
        return self._savepath

    @savepath.setter
    def savepath(self, value: str):
        self._savepath = value
        self.updateTitle()

    def save(self):
        """Save the current graph to the current save path."""
        self.scene.save(self.savepath)

    def saveAsJupyter(self, filepath: str):
        """Save the current graph notebook as a regular python notebook."""
        self.scene.save_to_ipynb(filepath)

    def load(self, filepath: str):
        """Load a graph from a file."""
        self.scene.load(filepath)
        self.savepath = filepath

    def moveToItems(self):
        """
        Ajust zoom and position to make the whole graph visible.
        If items are selected, then make all the selected items visible instead
        """
        self.view.moveToItems()



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for converting pygraph (.ipyg) data to notebook (.ipynb) data."""

from typing import OrderedDict, List

import copy

from pyflow.scene.ipynb_conversion_constants import *


def ipyg_to_ipynb(data: OrderedDict) -> OrderedDict:
    """Convert ipyg data (as ordered dict) into ipynb data (as ordered dict)."""
    ordered_data: OrderedDict = get_block_in_order(data)

    ipynb_data: OrderedDict = copy.deepcopy(DEFAULT_NOTEBOOK_DATA)

    for block_data in ordered_data["blocks"]:
        if block_data["block_type"] in BLOCK_TYPE_SUPPORTED_FOR_IPYG_TO_IPYNB:
            ipynb_data["cells"].append(block_to_ipynb_cell(block_data))

    return ipynb_data


def get_block_in_order(data: OrderedDict) -> OrderedDict:
    """Changes the order of the blocks from random to the naturel flow of the text."""

    # Not implemented yet
    return data


def block_to_ipynb_cell(block_data: OrderedDict) -> OrderedDict:
    """Convert a ipyg block into its corresponding ipynb cell."""
    if block_data["block_type"] == BLOCK_TYPE_TO_NAME["code"]:
        cell_data: OrderedDict = copy.deepcopy(DEFAULT_CODE_CELL)
        cell_data["source"] = split_lines_and_add_newline(block_data["source"])
        return cell_data
    if block_data["block_type"] == BLOCK_TYPE_TO_NAME["markdown"]:
        cell_data: OrderedDict = copy.deepcopy(DEFAULT_MARKDOWN_CELL)
        cell_data["source"] = split_lines_and_add_newline(block_data["text"])
        return cell_data

    raise ValueError(
        f"The block type {block_data['block_type']} is not supported but has been declared as such"
    )


def split_lines_and_add_newline(text: str) -> List[str]:
    """Split the text and add a \\n at the end of each line
    This is the jupyter notebook default formatting for source, outputs and text."""
    lines = text.split("\n")
    for i in range(len(lines) - 1):
        lines[i] += "\n"
    return lines



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>
# pylint:disable=unused-argument

""" Module for the Title block widget.

The Title is a modified QLineEdit for PyFlow purpose.
"""

import time
from typing import OrderedDict
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QFocusEvent, QFont, QMouseEvent
from PyQt5.QtWidgets import QLineEdit, QWidget

from pyflow.core.serializable import Serializable


class Title(QLineEdit, Serializable):
    """The title of an Block. Needs to be double clicked to interact."""

    def __init__(
        self,
        text: str,
        color: str = "white",
        font: str = "Ubuntu",
        size: int = 12,
        parent: QWidget = None,
    ):
        """Create a new title for an Block."""
        Serializable.__init__(self)
        QLineEdit.__init__(self, text, parent)
        self.clickTime = None
        self.init_ui(color, font, size)
        self.setReadOnly(True)
        self.setCursorPosition(0)

    def init_ui(self, color: str, font: str, size: int):
        """Apply the style given to the title."""
        self.color = color
        self.setStyleSheet(
            f"""
            QLineEdit {{
                color : {self.color};
                background-color: transparent;
                border:none;
            }}"""
        )
        self.setFont(QFont(font, size))

    def mousePressEvent(self, event: QMouseEvent):
        """
        Detect double clicks and single clicks are react accordingly by
        dispatching the event to the parent or the current widget
        """
        if self.clickTime is None or (
            self.isReadOnly() and time.time() - self.clickTime > 0.3
        ):
            self.parent().mousePressEvent(event)
        elif self.isReadOnly():
            self.mouseDoubleClickEvent(event)
            super().mousePressEvent(event)
        else:
            super().mousePressEvent(event)
        self.clickTime = time.time()

    def focusOutEvent(self, event: QFocusEvent):
        """The title is read-only when focused is lost."""
        self.setReadOnly(True)
        self.setCursorPosition(0)
        self.deselect()

    def mouseDoubleClickEvent(self, event: QMouseEvent):
        """Toggle readonly mode when double clicking."""
        self.setReadOnly(not self.isReadOnly())
        if not self.isReadOnly():
            self.setFocus(Qt.MouseFocusReason)

    def serialize(self) -> OrderedDict:
        """Return a serialized version of this widget."""
        return OrderedDict(
            [
                ("color", self.color),
                ("font", self.font().family()),
                ("size", self.font().pointSize()),
            ]
        )

    def deserialize(self, data: OrderedDict, hashmap: dict = None, restore_id=True):
        """Restore a title from serialized data."""
        if restore_id:
            self.id = data.get("id", id(self))
        self.init_ui(data["color"], data["font"], data["size"])



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the handling a scene history. """

from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from pyflow.scene import Scene


class SceneHistory:
    """Helper object to handle undo/redo operations on an Scene.

    Args:
        scene: Scene reference.
        max_stack: Maximum size of the history stack (number of available undo).

    """

    def __init__(self, scene: "Scene", max_stack: int = 50):
        self.scene = scene
        self.history_stack = []
        self.current = -1
        self.max_stack = max_stack

    def undo(self):
        """Undo the last action by moving the current stamp backward and restoring."""
        if len(self.history_stack) > 0 and self.current > 0:
            self.current -= 1
            self.restore()

    def redo(self):
        """Redo the last undone action by moving the current stamp forward and restoring."""
        if len(self.history_stack) > 0 and self.current + 1 < len(self.history_stack):
            self.current += 1
            self.restore()

    def checkpoint(self, description: str, set_modified=True):
        """Store a snapshot of the scene in the history stack.

        Args:
            description: Description given to this checkpoint.
            set_modified: Whether the scene should be considered modified.

        """
        history_stamp = {"description": description, "snapshot": self.scene.serialize()}
        self.store(history_stamp)
        if set_modified:
            self.scene.has_been_modified = True

    def store(self, data: Any):
        """Store new data in the history stack, updating current checkpoint.
        Remove data that would be forward in the history stack.

        Args:
            data: Data to store in the history stack.

        """
        if self.current + 1 < len(self.history_stack):
            self.history_stack = self.history_stack[0 : self.current + 1]

        self.history_stack.append(data)

        if len(self.history_stack) > self.max_stack:
            self.history_stack.pop(0)

        self.current = min(self.current + 1, len(self.history_stack) - 1)

    def restore(self):
        """Restore the scene using the snapshot pointed by current in the history stack."""
        if len(self.history_stack) >= 0 and self.current >= 0:
            stamp = self.history_stack[self.current]
            snapshot = stamp["snapshot"]
            self.scene.deserialize(snapshot)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the base Code Block."""

from typing import OrderedDict
from PyQt5.QtWidgets import QPushButton, QTextEdit
from PyQt5.QtGui import QPen, QColor

from ansi2html import Ansi2HTMLConverter

from pyflow.blocks.block import Block

from pyflow.blocks.executableblock import ExecutableBlock
from pyflow.core.pyeditor import PythonEditor

conv = Ansi2HTMLConverter()


class CodeBlock(ExecutableBlock):

    """
    Code Block

    Features an area to edit code as well as a panel to display the output.

    The following is always true:
    output_panel_height + source_panel_height + edge_size*2 + title_height == height

    """

    DEFAULT_DATA = {
        **Block.DEFAULT_DATA,
        "source": "",
    }
    MANDATORY_FIELDS = Block.MANDATORY_FIELDS

    def __init__(self, source: str = "", **kwargs):

        """
        Create a new CodeBlock.
        Initialize all the child widgets specific to this block type.

        """

        super().__init__(**kwargs)
        self.source_editor = PythonEditor(self)

        self._source = ""
        self._stdout = ""

        self.source = source

        self.output_panel_height = self.height / 3
        self._min_output_panel_height = 20
        self._min_source_editor_height = 20

        self.output_closed = True
        self._splitter_size = [1, 1]
        self._cached_stdout = ""
        self.has_been_run = False
        self.blocks_to_run = []

        self._pen_outlines = [
            QPen(QColor("#7F000000")),  # Idle
            QPen(QColor("#FF0000")),  # Running
            QPen(QColor("#00ff00")),  # Transmitting
        ]

        # Add output pannel
        self.output_panel = self.init_output_panel()
        self.run_button = self.init_run_button()
        self.run_all_button = self.init_run_all_button()

        # Add splitter between source_editor and panel
        self.splitter.addWidget(self.source_editor)
        self.splitter.addWidget(self.output_panel)

        self.title_left_offset = 3 * self.edge_size

        # Build root widget into holder
        self.holder.setWidget(self.root)

        self.update_all()  # Set the geometry of display and source_editor

    def init_output_panel(self):
        """Initialize the output display widget: QLabel."""
        output_panel = QTextEdit()
        output_panel.setReadOnly(True)
        output_panel.setFont(self.source_editor.font())
        return output_panel

    def init_run_button(self):
        """Initialize the run button."""
        run_button = QPushButton(">", self.root)
        run_button.move(int(self.edge_size), int(self.edge_size / 2))
        run_button.setFixedSize(int(3 * self.edge_size), int(3 * self.edge_size))
        run_button.clicked.connect(self.handle_run_left)
        return run_button

    def init_run_all_button(self):
        """Initialize the run all button."""
        run_all_button = QPushButton(">>", self.root)
        run_all_button.setFixedSize(int(3 * self.edge_size), int(3 * self.edge_size))
        run_all_button.clicked.connect(self.handle_run_right)
        run_all_button.raise_()

        return run_all_button

    def handle_run_right(self):
        """Called when the button for "Run All" was pressed."""
        if self.run_state != 0:
            self._interrupt_execution()
        else:
            self.run_right()

    def handle_run_left(self):
        """Called when the button for "Run Left" was pressed."""
        if self.run_state != 0:
            self._interrupt_execution()
        else:
            self.run_left()

    def run_code(self):
        """Run the code in the block."""

        # Reset stdout
        self._cached_stdout = ""

        # Set button text to ...
        self.run_button.setText("...")
        self.run_all_button.setText("...")

        super().run_code()  # actually run the code

    def execution_finished(self):
        super().execution_finished()
        self.run_button.setText(">")
        self.run_all_button.setText(">>")

    def update_title(self):
        """Change the geometry of the title widget."""
        self.title_widget.setGeometry(
            int(self.edge_size) + self.run_button.width(),
            int(self.edge_size / 2),
            int(self.width - self.edge_size * 3 - self.run_button.width()),
            int(self.title_widget.height()),
        )

    def update_output_panel(self):
        """Change the geometry of the output panel."""
        # Close output panel if no output
        if self.stdout == "":
            self.previous_splitter_size = self.splitter.sizes()
            self.output_closed = True
            self.splitter.setSizes([1, 0])

    def update_run_all_button(self):
        """Change the geometry of the run all button."""
        self.run_all_button.move(
            int(self.width - self.edge_size - self.run_button.width()),
            int(self.edge_size / 2),
        )

    def update_all(self):
        """Update the code block parts."""
        super().update_all()
        self.update_output_panel()
        self.update_run_all_button()

    @property
    def source(self) -> str:
        """Source code."""
        return self._source

    @source.setter
    def source(self, value: str):
        if value != self._source:
            # If text has changed, set self and all output blocks to not run
            output_blocks, _ = self.custom_bfs(self, reverse=True)
            for block in output_blocks:
                block.has_been_run = False
            self.has_been_run = False
            self.source_editor.setText(value)
            self._source = value

    @property
    def pen_outline(self) -> QPen:
        """The current pen used to draw the outline of the CodeBlock."""
        return self._pen_outlines[self.run_state]

    @property
    def stdout(self) -> str:
        """Access the content of the output panel of the block."""
        return self._stdout

    @stdout.setter
    def stdout(self, value: str):
        self._stdout = value
        if hasattr(self, "output_panel"):
            if value.startswith("<img>"):
                display_text = self.b64_to_html(value[5:])
            elif value.startswith("<div>"):
                display_text = value
            else:
                display_text = self.str_to_html(value)
            self.output_panel.setText(display_text)
            # If output panel is closed and there is output, open it
            if self.output_closed and value != "":
                self.output_closed = False
                self.splitter.setSizes(self._splitter_size)
            # If output panel is open and there is no output, close it
            elif not self.output_closed and value == "":
                self._splitter_size = self.splitter.sizes()
                self.output_closed = True
                self.splitter.setSizes([1, 0])

    @staticmethod
    def str_to_html(text: str) -> str:
        """Format text so that it's properly displayed by the code block."""
        # Remove carriage returns and backspaces
        text = text.replace("\x08", "")
        text = text.replace("\r", "")
        # Convert ANSI escape codes to HTML
        text = conv.convert(text)
        # Replace background color
        text = text.replace(
            "background-color: #000000", "background-color: transparent"
        )
        return text

    def handle_stdout(self, value: str):
        """Handle the stdout signal."""
        # If there is a new line
        # Save every line but the last one

        if value.find("\n") != -1:
            lines = value.split("\n")
            self._cached_stdout += "\n".join(lines[:-1]) + "\n"
            value = lines[-1]

        # Update the last line only
        self.stdout = self._cached_stdout + value

    @staticmethod
    def b64_to_html(image: str) -> str:
        """Transform a base64 encoded image into a html image."""
        return f'<img src="data:image/png;base64,{image}">'

    def handle_image(self, image: str):
        """Handle the image signal."""
        self.stdout = "<img>" + image

    def serialize(self):
        """Serialize the code block."""
        base_dict = super().serialize()
        base_dict["source"] = self.source
        base_dict["stdout"] = self.stdout

        return base_dict

    def deserialize(
        self, data: OrderedDict, hashmap: dict = None, restore_id: bool = True
    ):
        """Restore a codeblock from it's serialized state."""

        self.complete_with_default(data)

        for dataname in ("source", "stdout"):
            if dataname in data:
                setattr(self, dataname, data[dataname])
        super().deserialize(data, hashmap, restore_id)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the SliderBlock.

A block that can allows dynamic value modification using a slider.

"""

from typing import OrderedDict
from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QHBoxLayout, QLabel, QLineEdit, QSlider, QVBoxLayout
from pyflow.blocks.executableblock import ExecutableBlock


class SliderBlock(ExecutableBlock):
    """
    Features a slider ranging from 0 to 1 and an area to choose what value to assign the slider to.
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        self.layout = QVBoxLayout(self.root)

        self.slider = QSlider(Qt.Horizontal)

        self.variable_layout = QHBoxLayout()
        self.variable_text = QLineEdit("slider_value")
        self.variable_value = QLabel(f"{self.slider.value()/100}")

        self.variable_text.setFixedWidth(self.root.width() / 2)

        self.variable_layout.addWidget(self.variable_text)
        self.variable_layout.addWidget(self.variable_value)

        self.layout.setContentsMargins(
            self.edge_size * 2,
            self.title_widget.height() + self.edge_size * 2,
            self.edge_size * 2,
            self.edge_size * 2,
        )
        self.layout.addWidget(self.slider)
        self.layout.addLayout(self.variable_layout)

        self.slider.valueChanged.connect(self.valueChanged)

        self.holder.setWidget(self.root)

    def valueChanged(self):
        """This is called when the value of the slider changes."""
        self.variable_value.setText(f"{self.value}")
        # Make sure that the slider is initialized before trying to run it.
        if self.scene() is not None:
            self.run_right()

    @property
    def source(self):
        """The "source code" of the slider i.e an assignement to the value of the slider."""
        python_code = f"{self.var_name} = {self.value}"
        return python_code

    @source.setter
    def source(self, value: str):
        raise RuntimeError("The source of a sliderblock is read-only.")

    @property
    def value(self):
        """The value of the slider."""
        return str(self.slider.value() / 100)

    @value.setter
    def value(self, value: str):
        self.slider.setValue(int(float(value) * 100))

    @property
    def var_name(self):
        """The name of the python variable associated with the slider."""
        return self.variable_text.text()

    @var_name.setter
    def var_name(self, value: str):
        self.variable_text.setText(value)

    def serialize(self):
        """Return a serialized version of this widget."""
        base_dict = super().serialize()
        base_dict["value"] = self.value
        base_dict["var_name"] = self.var_name

        return base_dict

    def deserialize(
        self, data: OrderedDict, hashmap: dict = None, restore_id: bool = True
    ):
        """Restore a slider block from it's serialized state."""
        for dataname in ["value", "var_name"]:
            if dataname in data:
                setattr(self, dataname, data[dataname])

        super().deserialize(data, hashmap, restore_id)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the Scene creation and manipulations."""

from pyflow.scene.scene import Scene



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module with the constants used to convert to and from notebooks."""

from typing import Dict

MARGIN_X: float = 75
MARGIN_BETWEEN_BLOCKS_X: float = 50
MARGIN_Y: float = 60
MARGIN_BETWEEN_BLOCKS_Y: float = 5
BLOCK_MIN_WIDTH: float = 400
TITLE_MAX_LENGTH: int = 60
SOCKET_HEIGHT: float = 44.0

DEFAULT_LINE_SPACING = 2
DEFAULT_LINE_HEIGHT = 10
DEFAULT_TEXT_WIDTH = 618

BLOCK_TYPE_TO_NAME: Dict[str, str] = {
    "code": "CodeBlock",
    "markdown": "MarkdownBlock",
}

BLOCK_TYPE_SUPPORTED_FOR_IPYG_TO_IPYNB = {"CodeBlock", "MarkdownBlock"}

DEFAULT_NOTEBOOK_DATA = {
    "cells": [],
    "metadata": {
        "kernelspec": {
            "display_name": "Python 3",
            "language": "python",
            "name": "python3",
        },
        "language_info": {
            "codemirror_mode": {"name": "ipython", "version": 3},
            "file_extension": ".py",
            "mimetype": "text/x-python",
            "name": "python",
            "nbconvert_exporter": "python",
            "pygments_lexer": "ipython3",
            "version": "3.6.4",
        },
    },
    "nbformat": 4,
    "nbformat_minor": 4,
}

DEFAULT_CODE_CELL = {
    "cell_type": "code",
    "execution_count": None,
    "metadata": {
        "_cell_guid": "b1076dfc-b9ad-4769-8c92-a6c4dae69d19",
        "_uuid": "8f2839f25d086af736a60e9eeb907d3b93b6e0e5",
        "execution": {
            "iopub.execute_input": "2021-11-23T21:43:41.246727Z",
            "iopub.status.busy": "2021-11-23T21:43:41.246168Z",
            "iopub.status.idle": "2021-11-23T21:43:41.260389Z",
            "shell.execute_reply": "2021-11-23T21:43:41.260950Z",
            "shell.execute_reply.started": "2021-11-22T18:36:28.843251Z",
        },
        "tags": [],
    },
    "outputs": [],
    "source": [],
}

DEFAULT_MARKDOWN_CELL = {
    "cell_type": "markdown",
    "metadata": {
        "papermill": {
            "duration": 0,
            "end_time": "2021-11-23T21:43:55.202848",
            "exception": False,
            "start_time": "2021-11-23T21:43:55.174774",
            "status": "completed",
        },
        "tags": [],
    },
    "source": [],
}



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for Theme.

Theme is a class that contains the details of the coloring theme.

"""

import json
from PyQt5.Qsci import QsciLexerPython
from PyQt5.QtGui import QColor


class Theme:
    """Class holding the details of a specific theme."""

    def __init__(self, name: str, json_str: str = "{}"):
        """
        Create a new theme
        """
        json_obj = json.loads(json_str)
        known_properties = {
            "comment_color": "#797979",
            "string_color": "#CE9178",
            "function_color": "#DCDCAA",
            "keyword_color": "#569CD6",
            "classname_color": "#4EC9B0",
            "literal_color": "#7FB347",
            "operator_color": "#D8D8D8",
        }
        for (property_name, property_value) in known_properties.items():
            if property_name in json_obj:
                setattr(self, property_name, json_obj[property_name])
            else:
                setattr(self, property_name, property_value)
        self.name = name

    def apply_to_lexer(self, lexer: QsciLexerPython):
        """Make the given lexer follow the theme."""
        lexer.setDefaultPaper(QColor("#1E1E1E"))
        lexer.setDefaultColor(QColor("#D4D4D4"))

        string_types = [
            QsciLexerPython.SingleQuotedString,
            QsciLexerPython.DoubleQuotedString,
            QsciLexerPython.UnclosedString,
            QsciLexerPython.SingleQuotedFString,
            QsciLexerPython.TripleSingleQuotedString,
            QsciLexerPython.TripleDoubleQuotedString,
            QsciLexerPython.TripleSingleQuotedFString,
            QsciLexerPython.TripleDoubleQuotedFString,
        ]

        for string_type in string_types:
            lexer.setColor(QColor(self.string_color), string_type)

        lexer.setColor(QColor(self.function_color), QsciLexerPython.FunctionMethodName)
        lexer.setColor(QColor(self.keyword_color), QsciLexerPython.Keyword)
        lexer.setColor(QColor(self.classname_color), QsciLexerPython.ClassName)
        lexer.setColor(QColor(self.literal_color), QsciLexerPython.Number)
        lexer.setColor(QColor(self.operator_color), QsciLexerPython.Operator)
        lexer.setColor(QColor(self.comment_color), QsciLexerPython.CommentBlock)
        lexer.setColor(QColor(self.comment_color), QsciLexerPython.Comment)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for base Sockets."""

from __future__ import annotations
from typing import List, Optional, OrderedDict, TYPE_CHECKING
import math

from PyQt5.QtCore import QPoint, QPointF, QRectF
from PyQt5.QtGui import QBrush, QColor, QPainter, QPen, QPolygon
from PyQt5.QtWidgets import QGraphicsItem, QStyleOptionGraphicsItem, QWidget

from pyflow.core.serializable import Serializable

if TYPE_CHECKING:
    from pyflow.core.edge import Edge
    from pyflow.blocks.block import Block


class Socket(QGraphicsItem, Serializable):

    """Base class for sockets in Pyflow."""

    DEFAULT_DATA = {
        "type": "undefined",
        "metadata": {
            "color": "#FF55FFF0",
            "linecolor": "#FF000000",
            "linewidth": 1.0,
            "radius": 6.0,
        },
    }
    MANDATORY_FIELDS = {"position"}

    def __init__(
        self,
        block: "Block",
        socket_type: str = DEFAULT_DATA["type"],
        flow_type: str = "exe",
        radius: float = DEFAULT_DATA["metadata"]["radius"],
        color: str = DEFAULT_DATA["metadata"]["color"],
        linewidth: float = DEFAULT_DATA["metadata"]["linewidth"],
        linecolor: str = DEFAULT_DATA["metadata"]["linecolor"],
    ):
        """Base class for sockets in Pyflow.

        Args:
            block: Block containing the socket.
            socket_type: Type of the socket, one of ('undefined', 'input', 'output').
            flow_type: Type of flow that the socket is linked to, one of ('exe', 'var').
            radius: Radius of the socket graphics.
            color: Color of the socket graphics.
            linewidth: Linewidth of the socket graphics.
            linecolor: Linecolor of the socket graphics.

        """
        Serializable.__init__(self)
        self.block = block
        QGraphicsItem.__init__(self, parent=self.block)

        self.edges: List["Edge"] = []
        self.socket_type = socket_type
        self.flow_type = flow_type

        self.radius = radius
        self._pen = QPen(QColor(linecolor))
        self._pen.setWidth(int(linewidth))
        self._brush = QBrush(QColor(color))

        self.metadata = {
            "radius": radius,
            "color": color,
            "linewidth": linewidth,
            "linecolor": linecolor,
        }

    def add_edge(self, edge: "Edge", is_destination: bool):
        """Add a given edge to the socket edges."""
        if not self._allow_multiple_edges:
            for prev_edge in self.edges:
                prev_edge.remove()
        if self.flow_type == "exe":
            if (is_destination and self.socket_type != "input") or (
                not is_destination and self.socket_type != "output"
            ):
                edge.remove()
                return
        self.edges.append(edge)

    def remove_edge(self, edge: "Edge"):
        """Remove a given edge from the socket edges."""
        self.edges.remove(edge)

    def remove(self):
        """Remove the socket and all its edges from the scene it is in."""
        for edge in self.edges:
            edge.remove()
        scene = self.scene()
        if scene is not None:
            scene.removeItem(self)
        self.setParentItem(None)

    @property
    def _allow_multiple_edges(self):
        if self.flow_type == "exe":
            return True
        raise NotImplementedError

    def paint(
        self,
        painter: QPainter,
        option: QStyleOptionGraphicsItem,  # pylint:disable=unused-argument
        widget: Optional[QWidget] = None,
    ):  # pylint:disable=unused-argument
        """Paint the socket."""
        painter.setBrush(self._brush)
        painter.setPen(self._pen)
        r = self.radius
        if self.flow_type == "exe":
            angles = [0, 2 * math.pi / 3, -2 * math.pi / 3]
            right_triangle_points = [
                QPoint(int(r * math.cos(angle)), int(r * math.sin(angle)))
                for angle in angles
            ]
            painter.drawPolygon(QPolygon(right_triangle_points))
        else:
            painter.drawEllipse(int(-r), int(-r), int(2 * r), int(2 * r))

    def boundingRect(self) -> QRectF:
        """Get the socket bounding box."""
        r = self.radius
        return QRectF(-r, -r, 2 * r, 2 * r)

    def serialize(self) -> OrderedDict:
        metadata = OrderedDict(sorted(self.metadata.items()))
        return OrderedDict(
            [
                ("id", self.id),
                ("type", self.socket_type),
                ("position", [self.pos().x(), self.pos().y()]),
                ("metadata", metadata),
            ]
        )

    def deserialize(self, data: OrderedDict, hashmap: dict = None, restore_id=True):
        if restore_id and "id" in data:
            self.id = data["id"]

        self.complete_with_default(data)

        self.socket_type = data["type"]
        self.setPos(QPointF(*data["position"]))

        self.metadata = dict(data["metadata"])
        self._pen.setColor(QColor(self.metadata["linecolor"]))
        self._pen.setWidth(int(self.metadata["linewidth"]))
        self._brush.setColor(QColor(self.metadata["color"]))



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>
# pylint:disable=too-many-instance-attributes, unsubscriptable-object

""" Module for the base Window."""

import os
import pathlib

from PyQt5.QtCore import QPoint, QSettings, QSize, Qt, QSignalMapper
from PyQt5.QtGui import QCloseEvent, QKeySequence
from PyQt5.QtWidgets import (
    QWidget,
    QAction,
    QFileDialog,
    QMainWindow,
    QMessageBox,
    QMdiArea,
)

from pyflow.graphics.widget import Widget
from pyflow.graphics.theme_manager import theme_manager

from pyflow.qss import loadStylesheets
from pyflow.qss import __file__ as QSS_INIT_PATH

QSS_PATH = pathlib.Path(QSS_INIT_PATH).parent


class Window(QMainWindow):

    """Main window of the Pyflow Qt-based application."""

    def __init__(self):
        super().__init__()

        self.stylesheets = [
            os.path.join(QSS_PATH, "pyflow.qss"),
            os.path.join(QSS_PATH, "pyflow_dark.qss"),
        ]
        loadStylesheets(self.stylesheets)

        self.mdiArea = QMdiArea()
        self.mdiArea.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.mdiArea.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)
        self.mdiArea.setViewMode(QMdiArea.ViewMode.TabbedView)
        self.mdiArea.setDocumentMode(True)
        self.mdiArea.setTabsMovable(True)
        self.mdiArea.setTabsClosable(True)
        self.setCentralWidget(self.mdiArea)

        self.mdiArea.subWindowActivated.connect(self.updateMenus)
        self.windowMapper = QSignalMapper(self)
        self.windowMapper.mapped[QWidget].connect(self.setActiveSubWindow)

        self.themeMapper = QSignalMapper(self)
        self.themeMapper.mapped[int].connect(self.setTheme)

        # Menus
        self.createActions()
        self.createMenus()
        self.createToolBars()

        # StatusBar
        self.statusbar = self.statusBar()

        self.updateMenus()

        # Window properties
        self.readSettings()
        self.show()

    def createToolBars(self):
        """Does nothing, but is required by the QMainWindow interface."""

    def updateMenus(self):
        """Does nothing, but is required by the QMainWindow interface."""

    def createActions(self):
        """Create all menu actions."""
        # File
        self._actNew = QAction(
            "&New",
            statusTip="Create new ipygraph",
            shortcut="Ctrl+N",
            triggered=self.onFileNew,
        )
        self._actOpen = QAction(
            "&Open",
            statusTip="Open an ipygraph",
            shortcut="Ctrl+O",
            triggered=self.onFileOpen,
        )
        self._actSave = QAction(
            "&Save",
            statusTip="Save the ipygraph",
            shortcut="Ctrl+S",
            triggered=self.onFileSave,
        )
        self._actSaveAs = QAction(
            "Save &As...",
            statusTip="Save the ipygraph as...",
            shortcut="Ctrl+Shift+S",
            triggered=self.onFileSaveAs,
        )
        self._actSaveAsJupyter = QAction(
            "Save &As ... .ipynb",
            statusTip="Save the ipygraph as a Jupter Notebook at ...",
            triggered=self.onFileSaveAsJupyter,
        )
        self._actQuit = QAction(
            "&Quit",
            statusTip="Save and Quit the application",
            shortcut="Ctrl+Q",
            triggered=self.close,
        )

        # Edit
        self._actUndo = QAction(
            "&Undo",
            statusTip="Undo last operation",
            shortcut="Ctrl+Z",
            triggered=self.onEditUndo,
        )
        self._actRedo = QAction(
            "&Redo",
            statusTip="Redo last operation",
            shortcut="Ctrl+Y",
            triggered=self.onEditRedo,
        )
        self._actCut = QAction(
            "Cu&t",
            statusTip="Cut to clipboard",
            shortcut="Ctrl+X",
            triggered=self.onEditCut,
        )
        self._actCopy = QAction(
            "&Copy",
            statusTip="Copy to clipboard",
            shortcut="Ctrl+C",
            triggered=self.onEditCopy,
        )
        self._actPaste = QAction(
            "&Paste",
            statusTip="Paste from clipboard",
            shortcut="Ctrl+V",
            triggered=self.onEditPaste,
        )
        self._actDel = QAction(
            "&Del",
            statusTip="Delete selected items",
            shortcut="Del",
            triggered=self.onEditDelete,
        )

        # View
        self._actViewItems = QAction(
            "See All Blocks",
            statusTip="See all selected blocks. If none are selected, view all blocks",
            shortcut=" ",
            triggered=self.onMoveToItems,
        )

        # Window
        self._actClose = QAction(
            "Cl&ose",
            self,
            statusTip="Close the active window",
            triggered=self.mdiArea.closeActiveSubWindow,
        )
        self._actCloseAll = QAction(
            "Close &All",
            self,
            statusTip="Close all the windows",
            triggered=self.mdiArea.closeAllSubWindows,
        )
        self._actTile = QAction(
            "&Tile",
            self,
            statusTip="Tile the windows",
            triggered=self.mdiArea.tileSubWindows,
        )
        self._actCascade = QAction(
            "&Cascade",
            self,
            statusTip="Cascade the windows",
            triggered=self.mdiArea.cascadeSubWindows,
        )
        self._actNext = QAction(
            "Ne&xt",
            self,
            shortcut=QKeySequence.StandardKey.NextChild,
            statusTip="Move the focus to the next window",
            triggered=self.mdiArea.activateNextSubWindow,
        )
        self._actPrevious = QAction(
            "Pre&vious",
            self,
            shortcut=QKeySequence.StandardKey.PreviousChild,
            statusTip="Move the focus to the previous window",
            triggered=self.mdiArea.activatePreviousSubWindow,
        )
        self._actSeparator = QAction(self)
        self._actSeparator.setSeparator(True)

    def createMenus(self):
        """Create the File menu with linked shortcuts."""
        self.filemenu = self.menuBar().addMenu("&File")
        self.filemenu.addAction(self._actNew)
        self.filemenu.addAction(self._actOpen)
        self.filemenu.addSeparator()
        self.filemenu.addAction(self._actSave)
        self.filemenu.addAction(self._actSaveAs)
        self.filemenu.addAction(self._actSaveAsJupyter)
        self.filemenu.addSeparator()
        self.filemenu.addAction(self._actQuit)

        self.editmenu = self.menuBar().addMenu("&Edit")
        self.editmenu.addAction(self._actUndo)
        self.editmenu.addAction(self._actRedo)
        self.editmenu.addSeparator()
        self.editmenu.addAction(self._actCut)
        self.editmenu.addAction(self._actCopy)
        self.editmenu.addAction(self._actPaste)
        self.editmenu.addSeparator()
        self.editmenu.addAction(self._actDel)

        self.viewmenu = self.menuBar().addMenu("&View")
        self.thememenu = self.viewmenu.addMenu("Theme")
        self.thememenu.aboutToShow.connect(self.updateThemeMenu)
        self.viewmenu.addAction(self._actViewItems)

        self.windowMenu = self.menuBar().addMenu("&Window")
        self.updateWindowMenu()
        self.windowMenu.aboutToShow.connect(self.updateWindowMenu)

        self.menuBar().addSeparator()

    def updateThemeMenu(self):
        """Update the theme selection menu."""
        self.thememenu.clear()
        theme_names = theme_manager().list_themes()
        for i, theme in enumerate(theme_names):
            action = self.thememenu.addAction(theme)
            action.setCheckable(True)
            action.setChecked(i == theme_manager().selected_theme_index)
            action.triggered.connect(self.themeMapper.map)
            self.themeMapper.setMapping(action, i)

    def updateWindowMenu(self):
        """Update the window menu."""
        self.windowMenu.clear()
        self.windowMenu.addAction(self._actClose)
        self.windowMenu.addAction(self._actCloseAll)
        self.windowMenu.addSeparator()
        self.windowMenu.addAction(self._actTile)
        self.windowMenu.addAction(self._actCascade)
        self.windowMenu.addSeparator()
        self.windowMenu.addAction(self._actNext)
        self.windowMenu.addAction(self._actPrevious)
        self.windowMenu.addAction(self._actSeparator)

        windows = self.mdiArea.subWindowList()
        self._actSeparator.setVisible(len(windows) != 0)

        for i, window in enumerate(windows):
            child = window.widget()

            text = f"{i + 1} {child.windowTitle()}"
            if i < 9:
                text = "&" + text

            action = self.windowMenu.addAction(text)
            action.setCheckable(True)
            action.setChecked(child is self.activeMdiChild())
            action.triggered.connect(self.windowMapper.map)
            self.windowMapper.setMapping(action, window)

    def createNewMdiChild(self, filename: str = None):
        """Create a new graph subwindow loading a file if a path is given."""
        _widget = Widget()
        if filename is not None:
            _widget.scene.load(filename)
            if filename.split(".")[-1] == "ipyg":
                _widget.savepath = filename
        return self.mdiArea.addSubWindow(_widget)

    def onFileNew(self):
        """Create a new file."""
        subwnd = self.createNewMdiChild()
        subwnd.show()

    def onFileOpen(self):
        """Open a file."""
        filename, _ = QFileDialog.getOpenFileName(self, "Open ipygraph from file")
        if filename == "":
            return
        if os.path.isfile(filename):
            subwnd = self.createNewMdiChild(filename)
            subwnd.show()
            self.statusbar.showMessage(f"Successfully loaded {filename}", 2000)

    def onFileSave(self) -> bool:
        """Save file.

        Returns:
            True if the file was successfully saved, False otherwise.

        """
        current_window = self.activeMdiChild()

        if current_window is not None:
            if current_window.savepath is None:
                return self.onFileSaveAs()
            self.saveWindow(current_window)
        return True

    def onFileSaveAs(self) -> bool:
        """Save file in a given directory, caching savepath for quick save.

        Returns:
            True if the file was successfully saved, False otherwise.

        """
        current_window = self.activeMdiChild()
        if current_window is not None:
            dialog = QFileDialog()
            dialog.setDefaultSuffix(".ipyg")
            filename, _ = dialog.getSaveFileName(
                self, "Save ipygraph to file", filter="IPython Graph (*.ipyg)"
            )
            if filename == "":
                return False
            current_window.savepath = filename

            # Note : the current_window is the activeMdiChild before the QFileDialog pops up
            self.saveWindow(current_window)
            return True
        return False

    def onFileSaveAsJupyter(self) -> bool:
        """Save file in a given directory as ipynb, caching savepath for quick save.

        Returns:
            True if the file was successfully saved, False otherwise.

        """
        current_window = self.activeMdiChild()
        if current_window is not None:
            dialog = QFileDialog()
            dialog.setDefaultSuffix(".ipynb")
            filename, _ = dialog.getSaveFileName(
                self, "Save ipygraph to file", filter="Jupyter Notebook (*.ipynb)"
            )
            if filename == "":
                return False
            current_window.saveAsJupyter(filename)
            self.statusbar.showMessage(
                f"Successfully saved ipygraph as jupter notebook at {filename}",
                2000,
            )
            return True
        return False

    def saveWindow(self, window: Widget):
        """Save the given window."""
        window.save()
        self.statusbar.showMessage(
            f"Successfully saved ipygraph at {window.savepath}", 2000
        )

    @staticmethod
    def is_not_editing(current_window: Widget):
        """True if current_window exists and is not in editing mode."""
        return current_window is not None and not current_window.view.is_mode("EDITING")

    def onEditUndo(self):
        """Undo last operation if not in edit mode."""
        current_window = self.activeMdiChild()
        if self.is_not_editing(current_window):
            current_window.scene.history.undo()

    def onEditRedo(self):
        """Redo last operation if not in edit mode."""
        current_window = self.activeMdiChild()
        if self.is_not_editing(current_window):
            current_window.scene.history.redo()

    def onEditCut(self):
        """Cut the selected items if not in edit mode."""
        current_window = self.activeMdiChild()
        if self.is_not_editing(current_window):
            current_window.scene.clipboard.cut()

    def onEditCopy(self):
        """Copy the selected items if not in edit mode."""
        current_window = self.activeMdiChild()
        if self.is_not_editing(current_window):
            current_window.scene.clipboard.copy()

    def onEditPaste(self):
        """Paste the selected items if not in edit mode."""
        current_window = self.activeMdiChild()
        if self.is_not_editing(current_window):
            current_window.scene.clipboard.paste()

    def onEditDelete(self):
        """Delete the selected items if not in edit mode."""
        current_window = self.activeMdiChild()
        if self.is_not_editing(current_window):
            current_window.view.deleteSelected()

    # def closeEvent(self, event:QEvent):
    #     """ Save and quit the application. """
    #     if self.maybeSave():
    #         event.accept()
    #     else:
    #         event.ignore()

    def closeEvent(self, event: QCloseEvent):
        """Save and quit the application."""
        self.mdiArea.closeAllSubWindows()
        if self.mdiArea.currentSubWindow():
            event.ignore()
        else:
            self.writeSettings()
            event.accept()

    def maybeSave(self) -> bool:
        """Ask for save and returns if the file should be closed.

        Returns:
            True if the file should be closed, False otherwise.

        """
        if not self.isModified():
            return True

        answer = QMessageBox.warning(
            self,
            "About to loose you work?",
            "The file has been modified.\n" "Do you want to save your changes?",
            QMessageBox.StandardButton.Save
            | QMessageBox.StandardButton.Discard
            | QMessageBox.StandardButton.Cancel,
        )

        if answer == QMessageBox.StandardButton.Save:
            return self.onFileSave()
        if answer == QMessageBox.StandardButton.Discard:
            return True
        return False

    def activeMdiChild(self) -> Widget:
        """Get the active Widget if existing."""
        activeSubWindow = self.mdiArea.activeSubWindow()
        if activeSubWindow is not None:
            return activeSubWindow.widget()
        return None

    def readSettings(self):
        """Read the settings from the config file."""
        settings = QSettings("AutopIA", "Pyflow")
        pos = settings.value("pos", QPoint(200, 200))
        size = settings.value("size", QSize(400, 400))
        self.move(pos)
        self.resize(size)
        if settings.value("isMaximized", False) == "true":
            self.showMaximized()

    def writeSettings(self):
        """Write the settings to the config file."""
        settings = QSettings("AutopIA", "Pyflow")
        settings.setValue("pos", self.pos())
        settings.setValue("size", self.size())
        settings.setValue("isMaximized", self.isMaximized())

    def setActiveSubWindow(self, window):
        """Set the active subwindow to the given window."""
        if window:
            self.mdiArea.setActiveSubWindow(window)

    def onMoveToItems(self):
        """
        Ajust zoom and position to make the whole graph in the current window visible.
        If items are selected, then make all the selected items visible instead
        """
        current_window = self.activeMdiChild()
        if current_window is not None and isinstance(current_window, Widget):
            current_window.moveToItems()

    def setTheme(self, theme_index):
        """Set the theme of the application."""
        theme_manager().selected_theme_index = theme_index



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the PyFlow python editor."""

from typing import TYPE_CHECKING, List
from PyQt5.QtCore import Qt
from PyQt5.QtGui import (
    QFocusEvent,
    QFont,
    QFontMetrics,
    QColor,
    QMouseEvent,
    QWheelEvent,
)
from PyQt5.Qsci import QsciScintilla, QsciLexerPython
from pyflow.graphics.theme_manager import theme_manager

from pyflow.blocks.block import Block

if TYPE_CHECKING:
    from pyflow.graphics.view import View

POINT_SIZE = 11


class PythonEditor(QsciScintilla):

    """In-block python editor for Pyflow."""

    def __init__(self, block: Block):
        """In-block python editor for Pyflow.

        Args:
            block: Block in which to add the python editor widget.

        """
        super().__init__(None)
        self._mode = "NOOP"
        self.block = block

        self.update_theme()
        theme_manager().themeChanged.connect(self.update_theme)

        # Set caret
        self.setCaretForegroundColor(QColor("#D4D4D4"))

        # Indentation
        self.setAutoIndent(True)
        self.setTabWidth(4)
        self.setIndentationGuides(True)
        self.setIndentationsUseTabs(False)
        self.setBackspaceUnindents(True)

        # Disable horizontal scrollbar
        self.SendScintilla(QsciScintilla.SCI_SETHSCROLLBAR, 0)

        # # Add folding
        # self.setFolding(QsciScintilla.FoldStyle.CircledTreeFoldStyle, 1)
        # self.setFoldMarginColors(background_color, background_color)
        # self.setMarkerForegroundColor(foreground_color, 0)
        # self.setMarkerBackgroundColor(background_color, 0)

        # Add background transparency
        self.setStyleSheet("background:transparent")
        self.setAttribute(Qt.WidgetAttribute.WA_TranslucentBackground)
        self.setWindowFlags(Qt.WindowType.FramelessWindowHint)

    def update_theme(self):
        """Change the font and colors of the editor to match the current theme."""
        font = QFont()
        font.setFamily(theme_manager().recommended_font_family)
        font.setFixedPitch(True)
        font.setPointSize(POINT_SIZE)
        self.setFont(font)

        # Margin 0 is used for line numbers
        fontmetrics = QFontMetrics(font)
        foreground_color = QColor("#dddddd")
        background_color = QColor("#212121")
        self.setMarginsFont(font)
        self.setMarginWidth(2, fontmetrics.width("00") + 6)
        self.setMarginLineNumbers(2, True)
        self.setMarginsForegroundColor(foreground_color)
        self.setMarginsBackgroundColor(background_color)

        lexer = QsciLexerPython()
        theme_manager().current_theme().apply_to_lexer(lexer)
        lexer.setFont(font)
        self.setLexer(lexer)

    def views(self) -> List["View"]:
        """Get the views in which the python_editor is present."""
        return self.block.scene().views()

    def wheelEvent(self, event: QWheelEvent) -> None:
        """How PythonEditor handles wheel events."""
        if self.mode == "EDITING" and event.angleDelta().x() == 0:
            event.accept()
            return super().wheelEvent(event)

    @property
    def mode(self) -> int:
        """PythonEditor current mode."""
        return self._mode

    @mode.setter
    def mode(self, value: str):
        self._mode = value
        for view in self.views():
            view.set_mode(value)

    def mousePressEvent(self, event: QMouseEvent) -> None:
        """PythonEditor reaction to PyQt mousePressEvent events."""
        if event.buttons() & Qt.MouseButton.LeftButton:
            self.mode = "EDITING"
        return super().mousePressEvent(event)

    def focusOutEvent(self, event: QFocusEvent):
        """PythonEditor reaction to PyQt focusOut events."""
        self.mode = "NOOP"
        self.block.source = self.text()
        return super().focusOutEvent(event)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the  Blocks of different types. """

from pyflow.blocks.sliderblock import SliderBlock
from pyflow.blocks.codeblock import CodeBlock
from pyflow.blocks.markdownblock import MarkdownBlock
from pyflow.blocks.drawingblock import DrawingBlock
from pyflow.blocks.containerblock import ContainerBlock



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for converting notebook (.ipynb) data to pygraph (.ipyg) data."""

from typing import OrderedDict, List

from PyQt5.QtGui import QFontMetrics, QFont

from pyflow.scene.ipynb_conversion_constants import *
from pyflow.graphics.theme_manager import theme_manager
from pyflow.core.pyeditor import POINT_SIZE


def ipynb_to_ipyg(data: OrderedDict, use_theme_font: bool = True) -> OrderedDict:
    """
    Convert ipynb data (ipynb file, as ordered dict) into ipyg data (ipyg, as ordered dict)
    - use_theme_font: should the height of the blocks be computed based on the current
    font selected.
    """

    blocks_data: List[OrderedDict] = get_blocks_data(data, use_theme_font)
    edges_data: List[OrderedDict] = get_edges_data(blocks_data)

    return {
        "blocks": blocks_data,
        "edges": edges_data,
    }


def get_blocks_data(
    data: OrderedDict, use_theme_font: bool = True
) -> List[OrderedDict]:
    """
    Get the blocks corresponding to a ipynb file,
    Returns them in the ipyg ordered dict format
    """

    if "cells" not in data:
        return []

    # Get the font metrics to determine the size fo the blocks
    fontmetrics = None
    if use_theme_font:
        font = QFont()
        font.setFamily(theme_manager().recommended_font_family)
        font.setFixedPitch(True)
        font.setPointSize(POINT_SIZE)
        fontmetrics = QFontMetrics(font)

    blocks_data: List[OrderedDict] = []

    next_block_x_pos: float = 0
    next_block_y_pos: float = 0
    next_block_id = 0

    for cell in data["cells"]:
        if "cell_type" not in cell or cell["cell_type"] not in ["code", "markdown"]:
            pass
        else:
            block_type: str = cell["cell_type"]

            text: str = cell["source"]

            text_width = DEFAULT_TEXT_WIDTH
            if use_theme_font:
                text_width: float = (
                    max(fontmetrics.boundingRect(line).width() for line in text)
                    if len(text) > 0
                    else 0
                )
            block_width: float = max(text_width + MARGIN_X, BLOCK_MIN_WIDTH)

            lineSpacing = DEFAULT_LINE_SPACING
            lineHeight = DEFAULT_LINE_HEIGHT

            if use_theme_font:
                lineSpacing = fontmetrics.lineSpacing()
                lineHeight = fontmetrics.lineWidth()

            text_height: float = len(text) * (lineSpacing + lineHeight)
            block_height: float = text_height + MARGIN_Y

            block_data = {
                "id": next_block_id,
                "block_type": BLOCK_TYPE_TO_NAME[block_type],
                "width": block_width,
                "height": block_height,
                "position": [
                    next_block_x_pos,
                    next_block_y_pos,
                ],
                "sockets": [],
            }

            if block_type == "code":
                block_data["source"] = "".join(text)
                next_block_y_pos = 0
                next_block_x_pos += block_width + MARGIN_BETWEEN_BLOCKS_X

                if len(blocks_data) > 0 and is_title(blocks_data[-1]):
                    block_title: OrderedDict = blocks_data.pop()
                    block_data["title"] = block_title["text"]

                    # Revert position effect of the markdown block
                    block_data["position"] = block_title["position"]
            elif block_type == "markdown":
                block_data.update(
                    {
                        "text": "".join(text),
                    }
                )
                next_block_y_pos += block_height + MARGIN_BETWEEN_BLOCKS_Y

            blocks_data.append(block_data)
            next_block_id += 1

    adujst_markdown_blocks_width(blocks_data)

    return blocks_data


def is_title(block_data: OrderedDict) -> bool:
    """Checks if the block is a one-line markdown block which could correspond to a title."""
    if block_data["block_type"] != BLOCK_TYPE_TO_NAME["markdown"]:
        return False
    if "\n" in block_data["text"]:
        return False
    if not block_data["text"] or len(block_data["text"]) > TITLE_MAX_LENGTH:
        return False
    # Headings, quotes, bold or italic text are not considered to be headings
    if block_data["text"][0] in {"#", "*", "`"}:
        return False
    return True


def adujst_markdown_blocks_width(blocks_data: OrderedDict) -> None:
    """
    Modify the markdown blocks width (in place)
    For them to match the width of block of code below
    """
    i: int = len(blocks_data) - 1

    while i >= 0:
        if blocks_data[i]["block_type"] == BLOCK_TYPE_TO_NAME["code"]:
            block_width: float = blocks_data[i]["width"]
            i -= 1

            while (
                i >= 0
                and blocks_data[i]["block_type"] == BLOCK_TYPE_TO_NAME["markdown"]
            ):
                blocks_data[i]["width"] = block_width
                i -= 1
        else:
            i -= 1


def get_edges_data(blocks_data: OrderedDict) -> OrderedDict:
    """Add sockets to the blocks (in place) and returns the edge list."""
    code_blocks: List[OrderedDict] = [
        block
        for block in blocks_data
        if block["block_type"] == BLOCK_TYPE_TO_NAME["code"]
    ]
    edges_data: List[OrderedDict] = []

    greatest_block_id: int = 0
    if len(blocks_data) > 0:
        greatest_block_id = blocks_data[-1]["id"]

    for i in range(1, len(code_blocks)):
        socket_id_out = greatest_block_id + 2 * i + 2
        socket_id_in = greatest_block_id + 2 * i + 1
        code_blocks[i - 1]["sockets"].append(
            get_output_socket_data(socket_id_out, code_blocks[i - 1]["width"])
        )
        code_blocks[i]["sockets"].append(get_input_socket_data(socket_id_in))
        edges_data.append(
            get_edge_data(
                i,
                code_blocks[i - 1]["id"],
                socket_id_out,
                code_blocks[i]["id"],
                socket_id_in,
            )
        )
    return edges_data


def get_input_socket_data(socket_id: int) -> OrderedDict:
    """Returns the input socket's data with the corresponding id."""
    return {
        "id": socket_id,
        "type": "input",
        "position": [0.0, SOCKET_HEIGHT],
    }


def get_output_socket_data(socket_id: int, block_width: int) -> OrderedDict:
    """
    Returns the input socket's data with the corresponding id
    and at the correct relative position with respect to the block
    """
    return {
        "id": socket_id,
        "type": "output",
        "position": [block_width, SOCKET_HEIGHT],
    }


def get_edge_data(
    edge_id: int,
    edge_start_block_id: int,
    edge_start_socket_id: int,
    edge_end_block_id: int,
    edge_end_socket_id: int,
) -> OrderedDict:
    """Return the ordered dict corresponding to the given parameters."""
    return {
        "id": edge_id,
        "source": {"block": edge_start_block_id, "socket": edge_start_socket_id},
        "destination": {"block": edge_end_block_id, "socket": edge_end_socket_id},
    }



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for SizeGrip block widget.

The size grip is the little icon at the bottom right of a block that is used to
resize a block.
"""

from PyQt5.QtCore import QPoint
from PyQt5.QtWidgets import QGraphicsItem, QSizeGrip, QWidget
from PyQt5.QtGui import QMouseEvent


class SizeGrip(QSizeGrip):
    """A grip to resize a block."""

    def __init__(self, block: QGraphicsItem, parent: QWidget = None):
        """
        Constructor for BlockSizeGrip

        block is the QGraphicsItem holding the QSizeGrip.
        It's usually an Block
        """
        super().__init__(parent)
        self.mouseX = 0
        self.mouseY = 0
        self.block = block
        self.resizing = False

    def mousePressEvent(self, mouseEvent: QMouseEvent):
        """Start the resizing."""
        self.mouseX = mouseEvent.globalX()
        self.mouseY = mouseEvent.globalY()
        self.resizing = True

    def mouseReleaseEvent(
        self, mouseEvent: QMouseEvent
    ):  # pylint:disable=unused-argument
        """Stop the resizing."""
        self.resizing = False
        self.block.scene().history.checkpoint("Resized block", set_modified=True)

    @property
    def _zoom(self) -> float:
        """Returns how much the scene is."""
        return self.block.scene().views()[0].zoom

    def mouseMoveEvent(self, mouseEvent: QMouseEvent):
        """Performs resizing of the root widget."""
        transformed_pt1 = self.block.mapFromScene(QPoint(0, 0))
        transformed_pt2 = self.block.mapFromScene(QPoint(1, 1))

        pt = transformed_pt2 - transformed_pt1
        pt /= self._zoom

        delta_x = (mouseEvent.globalX() - self.mouseX) * pt.x()
        delta_y = (mouseEvent.globalY() - self.mouseY) * pt.y()
        # Here, we use globalx and globaly instead of x() and y().
        # This is because when using x() and y(), the mouse position is taken
        # relative to the grip, so if the grip moves, the deltaX and deltaY changes.
        # This creates a shaking effect when resizing. We use global to not
        # have this effect.
        new_width = max(self.block.width + int(delta_x), self.block.min_width)
        new_height = max(self.block.height + int(delta_y), self.block.min_height)

        self.parent().setGeometry(0, 0, new_width, new_height)
        self.block.update_all()

        self.mouseX = mouseEvent.globalX()
        self.mouseY = mouseEvent.globalY()



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>
# pylint:disable=wrong-import-position

""" Pyflow main module. """

import os
import sys
import asyncio

if os.name == "nt":  # If on windows
    asyncio.set_event_loop_policy(asyncio.WindowsSelectorEventLoopPolicy())

from PyQt5.QtWidgets import QApplication

from pyflow.graphics.window import Window
from pyflow import __version__

sys.path.insert(0, os.path.join(os.path.dirname(__file__), "..", ".."))

if __name__ == "__main__":
    app = QApplication(sys.argv)
    app.setStyle("Fusion")
    wnd = Window()
    if len(sys.argv) > 1:
        wnd.createNewMdiChild(sys.argv[1])

    wnd.setWindowTitle(f"Pyflow {__version__}")
    wnd.show()
    sys.exit(app.exec_())



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module to create and manage multi-threading workers."""

import asyncio
from PyQt5.QtCore import QObject, pyqtSignal, QRunnable


class WorkerSignals(QObject):
    """Defines the signals available from a running worker thread."""

    stdout = pyqtSignal(str)
    image = pyqtSignal(str)
    finished = pyqtSignal()
    finished_block = pyqtSignal()
    error = pyqtSignal()


class Worker(QRunnable):
    """Worker thread."""

    def __init__(self, kernel, code):
        """Initialize the worker object."""
        super().__init__()

        self.kernel = kernel
        self.code = code
        self.signals = WorkerSignals()

    async def run_code(self):
        """Run the code in the block."""
        # Execute the code
        self.kernel.client.execute(self.code)
        done = False
        # While the kernel sends messages
        while done is False:
            # Save kernel message and send it to the GUI
            output, output_type, done = self.kernel.update_output()
            if done is False:
                if output_type == "text":
                    self.signals.stdout.emit(output)
                elif output_type == "image":
                    self.signals.image.emit(output)
                elif output_type == "error":
                    self.signals.error.emit()
                    self.signals.stdout.emit(output)
        self.signals.finished.emit()

    def run(self):
        """Execute the run_code method asynchronously."""
        loop = asyncio.new_event_loop()
        asyncio.set_event_loop(loop)
        loop.run_until_complete(self.run_code())
        loop.close()



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for qss and styles. """

from typing import List

from PyQt5.QtCore import QFile
from PyQt5.QtWidgets import QApplication

from pyflow.qss import dark_resources


def loadStylesheets(filenames: List[str]):
    """Load the stylesheets from the given filenames."""
    styles = ""
    for filename in filenames:
        file = QFile(filename)
        file.open(QFile.ReadOnly | QFile.Text)
        stylesheet = file.readAll()
        styles += "\n" + str(stylesheet, encoding="utf-8")
    QApplication.instance().setStyleSheet(styles)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the handling of scene clipboard operations. """

from typing import TYPE_CHECKING, OrderedDict
from warnings import warn

import json
from PyQt5.QtWidgets import QApplication

from pyflow.core.edge import Edge

if TYPE_CHECKING:
    from pyflow.scene import Scene
    from pyflow.graphics.view import View


class SceneClipboard:

    """Helper object to handle clipboard operations on an Scene."""

    def __init__(self, scene: "Scene"):
        """Helper object to handle clipboard operations on an Scene.

        Args:
            scene: Scene reference.

        """
        self.scene = scene

    def cut(self):
        """Cut the selected items and put them into clipboard."""
        self._store(self._serializeSelected(delete=True))

    def copy(self):
        """Copy the selected items into clipboard."""
        self._store(self._serializeSelected(delete=False))

    def paste(self):
        """Paste the items in clipboard into the current scene."""
        self._deserializeData(self._gatherData())

    def _serializeSelected(self, delete=False) -> OrderedDict:
        selected_blocks, selected_edges = self.scene.sortedSelectedItems()
        selected_sockets = {}

        # Gather selected sockets
        for block in selected_blocks:
            for socket in block.sockets_in + block.sockets_out:
                selected_sockets[socket.id] = socket

        # Filter edges that are not fully connected to selected sockets
        for edge in selected_edges:
            if (
                edge.source_socket.id not in selected_sockets
                or edge.destination_socket.id not in selected_sockets
            ):
                selected_edges.remove(edge)

        data = OrderedDict(
            [
                ("blocks", [block.serialize() for block in selected_blocks]),
                ("edges", [edge.serialize() for edge in selected_edges]),
            ]
        )

        if delete:  # Remove selected items
            self.scene.views()[0].deleteSelected()

        return data

    def _find_bbox_center(self, blocks_data):
        xmin, xmax, ymin, ymax = 0, 0, 0, 0
        for block_data in blocks_data:
            x, y = block_data["position"]
            if x < xmin:
                xmin = x
            if x > xmax:
                xmax = x
            if y < ymin:
                ymin = y
            if y > ymax:
                ymax = y
        return (xmin + xmax) / 2, (ymin + ymax) / 2

    def _deserializeData(self, data: OrderedDict, set_selected=True):
        if data is None:
            return

        hashmap = {}

        view = self.scene.views()[0]
        mouse_pos = view.lastMousePos
        if set_selected:
            self.scene.clearSelection()

        # Finding pasting bbox center
        bbox_center_x, bbox_center_y = self._find_bbox_center(data["blocks"])
        offset_x, offset_y = (
            mouse_pos.x() - bbox_center_x,
            mouse_pos.y() - bbox_center_y,
        )

        # Create blocks
        for block_data in data["blocks"]:
            block = self.scene.create_block(block_data, hashmap, restore_id=False)
            if set_selected:
                block.setSelected(True)
            block.setPos(block.x() + offset_x, block.y() + offset_y)

        # Create edges
        for edge_data in data["edges"]:
            edge = Edge()
            edge.deserialize(edge_data, hashmap, restore_id=False)

            if set_selected:
                edge.setSelected(True)
            self.scene.addItem(edge)
            hashmap.update({edge_data["id"]: edge})

        self.scene.history.checkpoint(
            "Desiralized elements into scene", set_modified=True
        )

    def _store(self, data: OrderedDict):
        str_data = json.dumps(data, indent=4)
        QApplication.instance().clipboard().setText(str_data)

    def _gatherData(self) -> str:
        str_data = QApplication.instance().clipboard().text()
        try:
            return json.loads(str_data)
        except ValueError as valueerror:
            warn(f"Clipboard text could not be loaded into json data: {valueerror}")
            return



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>
# pylint:disable=unused-argument

""" Module for the base Block."""

from typing import TYPE_CHECKING, Optional, OrderedDict, Tuple, Union

from PyQt5.QtCore import QPointF, QRectF, Qt
from PyQt5.QtGui import QBrush, QPen, QColor, QPainter, QPainterPath
from PyQt5.QtWidgets import (
    QGraphicsItem,
    QGraphicsProxyWidget,
    QGraphicsSceneMouseEvent,
    QStyleOptionGraphicsItem,
    QWidget,
)

from pyflow.core.serializable import Serializable
from pyflow.core.socket import Socket
from pyflow.blocks.widgets import Splitter, SizeGrip, Title

if TYPE_CHECKING:
    from pyflow.scene.scene import Scene

BACKGROUND_COLOR = QColor("#E3212121")


class Block(QGraphicsItem, Serializable):

    """Base class for blocks in Pyflow."""

    DEFAULT_DATA = {
        "title": "New block",
        "splitter_pos": [0, 0],
        "width": 300,
        "height": 200,
        "metadata": {
            "title_metadata": {"color": "white", "font": "Ubuntu", "size": 10}
        },
        "sockets": [],
    }
    MANDATORY_FIELDS = {"block_type", "position"}

    def __init__(
        self,
        block_type: str = "base",
        position: tuple = (0, 0),
        width: int = DEFAULT_DATA["width"],
        height: int = DEFAULT_DATA["height"],
        edge_size: float = 10.0,
        title: Union[Title, str] = DEFAULT_DATA["title"],
        parent: Optional["QGraphicsItem"] = None,
    ):
        """Base class for blocks in Pyflow.

        Args:
            block_type: Block type.
            position: Block position in the scene.
            width: Block width.
            height: Block height.
            edge_size: Block edges size.
            title: Block title.
            parent: Parent of the block.

        """
        QGraphicsItem.__init__(self, parent=parent)
        Serializable.__init__(self)

        self.block_type = block_type
        self.setPos(QPointF(*position))
        self.sockets_in = []
        self.sockets_out = []

        self._pen_outline = QPen(QColor("#7F000000"))
        self._pen_outline_selected = QPen(QColor("#FFFFA637"))
        self._brush_background = QBrush(BACKGROUND_COLOR)

        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsSelectable)
        self.setFlag(QGraphicsItem.GraphicsItemFlag.ItemIsMovable)

        self.setAcceptHoverEvents(True)

        self.holder = QGraphicsProxyWidget(self)
        self.root = QWidget()
        self.root.setAttribute(Qt.WA_TranslucentBackground)
        self.root.setGeometry(0, 0, int(width), int(height))

        self.title_widget = Title(title, parent=self.root)
        self.title_widget.setAttribute(Qt.WA_TranslucentBackground)

        self.splitter = Splitter(self, Qt.Vertical, self.root)

        self.size_grip = SizeGrip(self, self.root)

        if type(self) == Block:
            # This has to be called at the end of the constructor of
            # every class inheriting this.
            self.holder.setWidget(self.root)

        self.edge_size = edge_size
        self.min_width = 300
        self.min_height = 100
        self.width = width
        self.height = height

        self.moved = False
        self.metadata = {}

    def scene(self) -> "Scene":
        """Get the current Scene containing the block."""
        return super().scene()

    def boundingRect(self) -> QRectF:
        """Get the the block bounding box."""
        return QRectF(0, 0, self.width, self.height).normalized()

    def paint(
        self,
        painter: QPainter,
        option: QStyleOptionGraphicsItem,
        widget: Optional[QWidget] = None,
    ):
        """Paint the block."""

        # content
        path_content = QPainterPath()
        path_content.setFillRule(Qt.FillRule.WindingFill)
        path_content.addRoundedRect(
            0, 0, self.width, self.height, self.edge_size, self.edge_size
        )
        painter.setPen(Qt.PenStyle.NoPen)
        painter.setBrush(self._brush_background)
        painter.drawPath(path_content.simplified())

        # outline
        path_outline = QPainterPath()
        path_outline.addRoundedRect(
            0, 0, self.width, self.height, self.edge_size, self.edge_size
        )
        painter.setPen(
            self._pen_outline_selected if self.isSelected() else self.pen_outline
        )
        painter.setBrush(Qt.BrushStyle.NoBrush)
        painter.drawPath(path_outline.simplified())

    def get_socket_pos(self, socket: Socket) -> Tuple[float]:
        """Get a socket position to place them on the block sides."""
        if socket.socket_type == "input":
            x = 0
            sockets = self.sockets_in
        else:
            x = self.width
            sockets = self.sockets_out

        y_offset = self.title_widget.height() + 2 * socket.radius
        if len(sockets) < 2:
            y = y_offset
        else:
            side_lenght = self.height - y_offset - 2 * socket.radius - self.edge_size
            y = y_offset + side_lenght * sockets.index(socket) / (len(sockets) - 1)
        return x, y

    def update_sockets(self):
        """Update the sockets positions."""
        for socket in self.sockets_in + self.sockets_out:
            socket.setPos(*self.get_socket_pos(socket))

    def add_socket(self, socket: Socket):
        """Add a socket to the block."""
        if socket.socket_type == "input":
            self.sockets_in.append(socket)
        else:
            self.sockets_out.append(socket)
        self.update_sockets()

    def remove_socket(self, socket: Socket):
        """Remove a socket from the block."""
        if socket.socket_type == "input":
            self.sockets_in.remove(socket)
        else:
            self.sockets_out.remove(socket)
        socket.remove()
        self.update_sockets()

    def mouseReleaseEvent(self, event: QGraphicsSceneMouseEvent):
        """Block reaction to a mouseReleaseEvent."""
        if self.moved:
            self.moved = False
            self.scene().history.checkpoint("Moved block", set_modified=True)
        super().mouseReleaseEvent(event)

    def mouseMoveEvent(self, event: QGraphicsSceneMouseEvent):
        """Block reaction to a mouseMoveEvent."""
        super().mouseMoveEvent(event)
        self.moved = True

    def remove(self):
        """Remove the block from the scene containing it."""
        scene = self.scene()
        for socket in self.sockets_in + self.sockets_out:
            self.remove_socket(socket)
        if scene is not None:
            scene.removeItem(self)

    def update_splitter(self):
        """Change the geometry of the splitter to match the block."""
        # We make the resizing of splitter only affect
        # the last element of the split view
        self.splitter.setGeometry(
            int(self.edge_size),
            int(self.edge_size + self.title_widget.height()),
            int(self.width - self.edge_size * 2),
            int(self.height - self.edge_size * 2 - self.title_widget.height()),
        )

    def update_title(self):
        """Change the geometry of the title to match the block."""
        self.title_widget.setGeometry(
            int(self.edge_size),
            int(self.edge_size / 2),
            int(self.width - self.edge_size * 3),
            int(self.title_widget.height()),
        )

    def update_size_grip(self):
        """Change the geometry of the size grip to match the block."""
        self.size_grip.setGeometry(
            int(self.width - self.edge_size * 2),
            int(self.height - self.edge_size * 2),
            int(self.edge_size * 1.7),
            int(self.edge_size * 1.7),
        )

    def update_all(self):
        """Update sockets and title."""
        self.update_sockets()
        self.update_splitter()
        self.update_title()
        self.update_size_grip()

    @property
    def title(self):
        """Block title."""
        return self.title_widget.text()

    @title.setter
    def title(self, value: str):
        if hasattr(self, "title_widget"):
            self.title_widget.setText(value)

    @property
    def width(self):
        """Block width."""
        return self.root.width()

    @width.setter
    def width(self, value: float):
        self.root.setGeometry(0, 0, int(value), self.root.height())

    @property
    def height(self):
        """Block height."""
        return self.root.height()

    @height.setter
    def height(self, value: float):
        self.root.setGeometry(0, 0, self.root.width(), int(value))

    @property
    def pen_outline(self) -> QPen:
        """The current pen used to draw the outline of the Block."""
        return self._pen_outline

    def serialize(self) -> OrderedDict:
        """Return a serialized version of this widget."""
        self.metadata.update({"title_metadata": self.title_widget.serialize()})
        metadata = OrderedDict(sorted(self.metadata.items()))
        return OrderedDict(
            [
                ("id", self.id),
                ("title", self.title),
                ("block_type", self.block_type),
                ("splitter_pos", self.splitter.sizes()),
                ("position", [self.pos().x(), self.pos().y()]),
                ("width", self.width),
                ("height", self.height),
                ("metadata", metadata),
                (
                    "sockets",
                    [
                        socket.serialize()
                        for socket in self.sockets_in + self.sockets_out
                    ],
                ),
            ]
        )

    def deserialize(self, data: dict, hashmap: dict = None, restore_id=True) -> None:
        """Restore the block from serialized data."""
        if restore_id and "id" in data:
            self.id = data["id"]

        self.complete_with_default(data)

        for dataname in ("title", "block_type", "width", "height"):
            setattr(self, dataname, data[dataname])

        self.setPos(QPointF(*data["position"]))
        self.metadata = dict(data["metadata"])
        self.title_widget.deserialize(
            self.metadata["title_metadata"], hashmap, restore_id
        )

        if "splitter_pos" in data:
            self.splitter.setSizes(data["splitter_pos"])

        if len(data["sockets"]) > 0:
            # Remove old sockets
            for socket in self.sockets_in + self.sockets_out:
                self.remove_socket(socket)

            # Deserialize new sockets
            for socket_data in data["sockets"]:
                socket = Socket(block=self)
                socket.deserialize(socket_data, hashmap, restore_id)
                self.add_socket(socket)
                if hashmap is not None:
                    hashmap.update({socket_data["id"]: socket})

        self.update_all()



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the base Scene."""

import math
import json
from os import path
from types import FunctionType, ModuleType
from typing import List, OrderedDict, Union

from PyQt5.QtCore import QLine, QRectF, QThreadPool
from PyQt5.QtGui import QColor, QPainter, QPen
from PyQt5.QtWidgets import QGraphicsScene

from pyflow.core.serializable import Serializable
from pyflow.blocks.block import Block
from pyflow.core.edge import Edge
from pyflow.scene.clipboard import SceneClipboard
from pyflow.scene.history import SceneHistory
from pyflow.core.kernel import Kernel
from pyflow.scene.from_ipynb_conversion import ipynb_to_ipyg
from pyflow.scene.to_ipynb_conversion import ipyg_to_ipynb
from pyflow import blocks


class Scene(QGraphicsScene, Serializable):

    """Scene for the  Window."""

    def __init__(
        self,
        parent=None,
        background_color: str = "#393939",
        grid_color: str = "#292929",
        grid_light_color: str = "#2f2f2f",
        width: int = 64000,
        height: int = 64000,
        grid_size: int = 20,
        grid_squares: int = 5,
    ):
        Serializable.__init__(self)
        QGraphicsScene.__init__(self, parent=parent)

        self._background_color = QColor(background_color)
        self._grid_color = QColor(grid_color)
        self._grid_light_color = QColor(grid_light_color)
        self.grid_size = grid_size
        self.grid_squares = grid_squares

        self.width, self.height = width, height
        self.setSceneRect(-self.width // 2, -self.height // 2, self.width, self.height)
        self.setBackgroundBrush(self._background_color)

        self._has_been_modified = False
        self._has_been_modified_listeners = []

        self.history = SceneHistory(self)
        self.clipboard = SceneClipboard(self)

        self.kernel = Kernel()
        self.threadpool = QThreadPool()

    @property
    def has_been_modified(self):
        """True if the scene has been modified, False otherwise."""
        return self._has_been_modified

    @has_been_modified.setter
    def has_been_modified(self, value: bool):
        self._has_been_modified = value
        for callback in self._has_been_modified_listeners:
            callback()

    def addHasBeenModifiedListener(self, callback: FunctionType):
        """Add a callback that will trigger when the scene has been modified."""
        self._has_been_modified_listeners.append(callback)

    def sortedSelectedItems(self) -> List[Union[Block, Edge]]:
        """Returns the selected blocks and selected edges in two separate lists."""
        selected_blocks, selected_edges = [], []
        for item in self.selectedItems():
            if isinstance(item, Block):
                selected_blocks.append(item)
            if isinstance(item, Edge):
                selected_edges.append(item)
        return selected_blocks, selected_edges

    def drawBackground(self, painter: QPainter, rect: QRectF):
        """Draw the Scene background."""
        super().drawBackground(painter, rect)
        self.drawGrid(painter, rect)

    def drawGrid(self, painter: QPainter, rect: QRectF):
        """Draw the background grid."""
        left = int(math.floor(rect.left()))
        top = int(math.floor(rect.top()))
        right = int(math.ceil(rect.right()))
        bottom = int(math.ceil(rect.bottom()))

        first_left = left - (left % self.grid_size)
        first_top = top - (top % self.grid_size)

        # Compute all lines to be drawn
        lines_light, lines_dark = [], []
        for x in range(first_left, right, self.grid_size):
            if x % (self.grid_size * self.grid_squares) != 0:
                lines_light.append(QLine(x, top, x, bottom))
            else:
                lines_dark.append(QLine(x, top, x, bottom))

        for y in range(first_top, bottom, self.grid_size):
            if y % (self.grid_size * self.grid_squares) != 0:
                lines_light.append(QLine(left, y, right, y))
            else:
                lines_dark.append(QLine(left, y, right, y))

        # Draw the lines using the painter
        pen = QPen(self._grid_color)
        pen.setWidth(2)
        painter.setPen(pen)
        painter.drawLines(*lines_dark)

        pen = QPen(self._grid_light_color)
        pen.setWidth(1)
        painter.setPen(pen)
        painter.drawLines(*lines_light)

    def save(self, filepath: str):
        """Save the scene into filepath."""
        self.save_to_ipyg(filepath)
        self.has_been_modified = False

    def save_to_ipyg(self, filepath: str):
        """Save the scene into filepath as interactive python graph (.ipyg)."""
        if "." not in filepath:
            filepath += ".ipyg"

        extention_format = filepath.split(".")[-1]
        if extention_format != "ipyg":
            raise NotImplementedError(f"Unsupported format {extention_format}")

        with open(filepath, "w", encoding="utf-8") as file:
            file.write(json.dumps(self.serialize(), indent=4))

    def save_to_ipynb(self, filepath: str):
        """Save the scene into filepath as ipynb."""
        if "." not in filepath:
            filepath += ".ipynb"

        extention_format: str = filepath.split(".")[-1]
        if extention_format != "ipynb":
            raise NotImplementedError(
                f"The file should be a *.ipynb (not a .{extention_format})"
            )

        with open(filepath, "w", encoding="utf-8") as file:
            json_ipyg_data: OrderedDict = self.serialize()
            json_ipynb_data: OrderedDict = ipyg_to_ipynb(json_ipyg_data)
            file.write(json.dumps(json_ipynb_data, indent=4))

    def load(self, filepath: str):
        """Load a saved scene.

        Args:
            filepath: Path to the file to load.

        """
        if filepath.endswith(".ipyg"):
            data = self.load_from_json(filepath)
        elif filepath.endswith(".ipynb"):
            ipynb_data = self.load_from_json(filepath)
            data = ipynb_to_ipyg(ipynb_data)
        else:
            extention_format = filepath.split(".")[-1]
            raise NotImplementedError(f"Unsupported format {extention_format}")
        self.deserialize(data)
        self.history.checkpoint("Loaded scene")
        self.has_been_modified = False

        # Add filepath to kernel path
        dir_path = repr(path.abspath(path.dirname(filepath)))
        setup_path_code = f'__import__("os").chdir({dir_path})'
        self.kernel.execute(setup_path_code)

    def load_from_json(self, filepath: str) -> OrderedDict:
        """
        Load the json data into an ordered dict

        Args:
            filepath: Path to the file to load.
        """
        with open(filepath, "r", encoding="utf-8") as file:
            data = json.loads(file.read())
        return data

    def clear(self):
        """Clear the scene from all items."""
        self.has_been_modified = False
        return super().clear()

    def serialize(self) -> OrderedDict:
        """Serialize the scene into a dict."""
        blocks = []
        edges = []
        for item in self.items():
            if isinstance(item, Block):
                blocks.append(item)
            elif isinstance(item, Edge):
                edges.append(item)
        blocks.sort(key=lambda x: x.id)
        edges.sort(key=lambda x: x.id)
        return OrderedDict(
            [
                ("id", self.id),
                ("blocks", [block.serialize() for block in blocks]),
                ("edges", [edge.serialize() for edge in edges]),
            ]
        )

    def create_block_from_file(self, filepath: str, x: float = 0, y: float = 0):
        """Create a new block from a .b file."""
        with open(filepath, "r", encoding="utf-8") as file:
            data = json.loads(file.read())
            data["position"] = [x, y]
            data["sockets"] = {}
            self.create_block(data, None, False)

    def create_block(
        self, data: OrderedDict, hashmap: dict = None, restore_id: bool = True
    ) -> Block:
        """Create a new block from an OrderedDict."""

        block = None

        block_constructor = None
        block_files = blocks.__dict__

        for block_name in block_files:
            block_module = getattr(blocks, block_name)
            if isinstance(block_module, ModuleType):
                if hasattr(block_module, data["block_type"]):
                    block_constructor = getattr(blocks, data["block_type"])

        if block_constructor is None:
            raise NotImplementedError(f"{data['block_type']} is not a known block type")

        block = block_constructor()
        block.deserialize(data, hashmap, restore_id)
        self.addItem(block)
        if hashmap is not None:
            hashmap.update({data["id"]: block})
        return block

    def deserialize(
        self, data: OrderedDict, hashmap: dict = None, restore_id: bool = True
    ):
        self.clear()
        hashmap = hashmap if hashmap is not None else {}
        if restore_id and "id" in data:
            self.id = data["id"]

        # Create blocks
        for block_data in data["blocks"]:
            self.create_block(block_data, hashmap, restore_id)

        # Create edges
        for edge_data in data["edges"]:
            edge = Edge()
            edge.deserialize(edge_data, hashmap, restore_id)
            self.addItem(edge)
            hashmap.update({edge_data["id"]: edge})



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the MarkdownBlock.

A block able to render Markdown.

"""

from typing import OrderedDict
from markdown import markdown

from PyQt5.QtWebEngineWidgets import QWebEngineView
from PyQt5.Qsci import QsciLexerMarkdown, QsciScintilla
from PyQt5.QtCore import Qt
from PyQt5.QtGui import QColor, QFont
from pyflow.blocks.block import Block
from pyflow.graphics.theme_manager import theme_manager


class MarkdownBlock(Block):
    """A block that is able to render markdown text."""

    def __init__(self, **kwargs):
        """
        Create a new MarkdownBlock, a block that renders markdown
        """
        super().__init__(**kwargs)

        self.editor = QsciScintilla()
        self.editor.SendScintilla(QsciScintilla.SCI_SETHSCROLLBAR, 0)
        self.editor.setWindowFlags(Qt.WindowType.FramelessWindowHint)
        self.editor.setAutoFillBackground(False)

        self.lexer = QsciLexerMarkdown()
        theme_manager().current_theme().apply_to_lexer(self.lexer)
        self.lexer.setColor(QColor.fromRgb(255, 255, 255), -1)
        self.editor.setCaretForegroundColor(QColor("#FFFFFF"))
        self.editor.setLexer(self.lexer)

        font = QFont()
        font.setFamily(theme_manager().recommended_font_family)
        font.setFixedPitch(True)
        font.setPointSize(11)
        self.editor.setFont(font)
        self.editor.setMarginWidth(QsciScintilla.SC_MARGIN_NUMBER, 0)
        self.editor.setStyleSheet("background:transparent")
        self.editor.textChanged.connect(self.valueChanged)

        self.splitter.addWidget(self.editor)

        self.rendered_markdown = QWebEngineView()
        self.rendered_markdown.page().setBackgroundColor(
            QColor.fromRgba64(0, 0, 0, alpha=0)
        )

        self.splitter.addWidget(self.rendered_markdown)
        self.holder.setWidget(self.root)

    def valueChanged(self):
        """Update markdown rendering when the content of the markdown editor changes."""
        t = self.editor.text()

        dark_theme = """
            <style>
                *{
                    background-color:transparent;
                    color:white;
                }
            </style>
        """

        self.rendered_markdown.setHtml(f"{dark_theme}{markdown(t)}")

    @property
    def text(self) -> str:
        """The content of the markdown block."""
        return self.editor.text()

    @text.setter
    def text(self, value: str):
        self.editor.setText(value)
        self.valueChanged()

    def serialize(self):
        base_dict = super().serialize()
        base_dict["text"] = self.text

        return base_dict

    def deserialize(
        self, data: OrderedDict, hashmap: dict = None, restore_id: bool = True
    ):
        """Restore a markdown block from it's serialized state."""
        for dataname in ["text"]:
            if dataname in data:
                setattr(self, dataname, data[dataname])

        super().deserialize(data, hashmap, restore_id)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module to create and manage ipython kernels."""

import queue
from typing import Tuple
from jupyter_client.manager import start_new_kernel

from pyflow.core.worker import Worker


class Kernel:

    """jupyter_client kernel used to execute code and return output."""

    def __init__(self):
        self.kernel_manager, self.client = start_new_kernel()
        self.execution_queue = []
        self.busy = False

    def message_to_output(self, message: dict) -> Tuple[str, str]:
        """
        Converts a message sent by the kernel into a relevant output

        Args:
            message: dict representing the a message sent by the kernel

        Return:
            single output found in the message in that order of priority:
                image > text data > text print > error > nothing
        """
        message_type = "None"
        if "data" in message:
            if "image/png" in message["data"]:
                message_type = "image"
                # output an image (from plt.plot or plt.imshow)
                out = message["data"]["image/png"]
            elif "text/html" in message["data"]:
                message_type = "text"
                # output some html text (like a pandas dataframe)
                out = message["data"]["text/html"]
            else:
                message_type = "text"
                # output data as str (for example if code="a=10\na")
                out = message["data"]["text/plain"]
        elif "name" in message and message["name"] == "stdout":
            message_type = "text"
            # output a print (print("Hello World"))
            out = message["text"]
        elif "traceback" in message:
            message_type = "error"
            # output an error
            out = "\n".join(message["traceback"])
        else:
            message_type = "text"
            out = ""
        return out, message_type

    def run_block(self, block, code: str):
        """
        Runs code on a separate thread and sends the output to the block
        Also calls run_queue when finished

        Args:
            block: CodeBlock to send the output to
            code: String representing a piece of Python code to execute
        """
        worker = Worker(self, code)
        # Change color to running
        block.run_state = 1
        worker.signals.stdout.connect(block.handle_stdout)
        worker.signals.image.connect(block.handle_image)
        worker.signals.finished.connect(self.run_queue)
        worker.signals.finished.connect(block.execution_finished)
        worker.signals.error.connect(block.reset_has_been_run)
        block.scene().threadpool.start(worker)

    def run_queue(self):
        """Runs the next code in the queue."""
        self.busy = True
        if not self.execution_queue:
            self.busy = False
            return None
        block, code = self.execution_queue.pop(0)
        self.run_block(block, code)

    def execute(self, code: str) -> str:
        """
        Executes code in the kernel and returns the output of the last message sent by the kernel

        Args:
            code: String representing a piece of Python code to execute

        Return:
            output from the last message sent by the kernel
        """
        _ = self.client.execute(code)
        done = False
        while not done:
            # Check for messages, break the loop when the kernel stops sending messages
            new_message, done = self.get_message()
            if not done:
                message = new_message
        return self.message_to_output(message)[0]

    def get_message(self) -> Tuple[str, bool]:
        """
        Get message in the jupyter kernel

        Args:
            code: String representing a piece of Python code to execute

        Return:
            Tuple of:
                - output from the last message sent by the kernel
                - boolean repesenting if the kernel as any other message to send.
        """
        done = False
        try:
            message = self.client.get_iopub_msg(timeout=1000)["content"]
            if "execution_state" in message and message["execution_state"] == "idle":
                done = True
        except queue.Empty:
            message = None
            done = True
        return message, done

    def update_output(self) -> Tuple[str, str, bool]:
        """
        Returns the current output of the kernel

        Return:
            current output of the kernel; done: bool, True if the kernel has no message to send
        """
        message, done = self.get_message()
        out, output_type = self.message_to_output(message)
        return out, output_type, done

    def __del__(self):
        """
        Shuts down the kernel
        """
        self.kernel_manager.shutdown_kernel()



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

"""Setup for the python package build."""

import pathlib
from setuptools import setup, find_packages


def get_version():
    """Load version from file."""
    version_file = open("VERSION")
    return version_file.read().strip()


def get_requirements():
    """Load requirements from file."""
    requirements_file = open("requirements.txt")
    return requirements_file.readlines()


HERE = pathlib.Path(__file__).parent  # The directory containing this file
README = (HERE / "pypi-readme.md").read_text()
VERSION = get_version()
REQUIREMENTS = get_requirements()

setup(
    name="byc-pyflow",
    version=VERSION,
    author="Bycelium",
    author_email="mathis.federico@bycelium.com",
    description="An open-source tool for modular visual programing in python",
    long_description=README,
    long_description_content_type="text/markdown",
    url="https://github.com/Bycelium/PyFlow",
    packages=find_packages(exclude=["*test*", "*docs*"]),
    include_package_data=True,
    install_requires=REQUIREMENTS,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.7",
)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the ThemeManager.

This module provides `theme_manager()`,
a method that returns a handle to the theme manager of the application.

The theme manager provides the color scheme for the syntax highlighting
of the text areas containing code.
"""
import os
import pathlib
from typing import List

from PyQt5.QtGui import QFontDatabase
from PyQt5.QtCore import pyqtSignal, QObject

from pyflow.graphics.theme import Theme
from pyflow import __file__ as INIT_PATH

PACKAGE_PATH = pathlib.Path(INIT_PATH).parent


class ThemeManager(QObject):
    """Class loading theme files and providing the options set in those files."""

    themeChanged = pyqtSignal()

    def __init__(self, parent=None):
        """Load the default themes and the fonts available to construct the ThemeManager."""
        super().__init__(parent)
        self._preferred_fonts = ["Inconsolata", "Roboto Mono", "Courier"]
        self.recommended_font_family = "Monospace"
        qfd = QFontDatabase()
        available_fonts = qfd.families()
        for font in self._preferred_fonts:
            if font in available_fonts:
                self.recommended_font_family = font
                break

        self._themes = []
        self._selected_theme_index = 0
        theme_path = os.path.join(PACKAGE_PATH, "themes")
        theme_paths = os.listdir(theme_path)
        for p in theme_paths:
            full_path = os.path.join(theme_path, p)
            if os.path.isfile(full_path) and full_path.endswith(".theme"):
                name = os.path.splitext(os.path.basename(p))[0]
                with open(full_path, "r", encoding="utf-8") as f:
                    theme = Theme(name, f.read())
                    self._themes.append(theme)

    @property
    def selected_theme_index(self):
        """Return the index of the selected theme."""
        return self._selected_theme_index

    @selected_theme_index.setter
    def selected_theme_index(self, value: int):
        self._selected_theme_index = value
        self.themeChanged.emit()

    def list_themes(self) -> List[str]:
        """List the themes."""
        return [theme.name for theme in self._themes]

    def current_theme(self) -> Theme:
        """Return the current theme."""
        return self._themes[self.selected_theme_index]


theme_handle = None


def theme_manager():
    """Retreive the theme manager of the application."""
    global theme_handle
    if theme_handle is None:
        theme_handle = ThemeManager()
    return theme_handle



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the Splitter block widget.

The Splitter contains multiple areas inside a block
and allows the user to resize those areas.
"""

from PyQt5.QtGui import QMouseEvent
from PyQt5.QtWidgets import QSplitter, QSplitterHandle, QWidget


class SplitterHandle(QSplitterHandle):
    """A handle for splitters with undoable events."""

    def mouseReleaseEvent(self, evt: QMouseEvent):
        """When releasing the handle, save the state to history."""
        scene = self.parent().block.scene()
        if scene is not None:
            scene.history.checkpoint("Resize block", set_modified=True)
        return super().mouseReleaseEvent(evt)


class Splitter(QSplitter):
    """A spliter with undoable events."""

    def __init__(self, block: QWidget, orientation: int, parent: QWidget):
        """Create a new Splitter."""
        super().__init__(orientation, parent)
        self.block = block

    def createHandle(self):
        """Return the middle handle of the splitter."""
        return SplitterHandle(self.orientation(), self)



# Pyflow an open-source tool for modular visual programing in python
# Copyright (C) 2021-2022 Bycelium <https://www.gnu.org/licenses/>

""" Module for the Serializable base class """

from typing import Any, Dict, OrderedDict


class Serializable:

    """Serializable base for serializable objects."""

    MANDATORY_FIELDS: OrderedDict = {}
    DEFAULT_DATA: Dict[str, Any] = {}

    def __init__(self):
        self.id = id(self)

    def serialize(self) -> OrderedDict:
        """Serialize the object as an ordered dictionary."""
        raise NotImplementedError()

    def deserialize(
        self, data: OrderedDict, hashmap: dict = None, restore_id=True
    ) -> None:
        """Deserialize the object from an ordered dictionary.

        Args:
            data: Dictionnary containing data do deserialize from.
            hashmap: Dictionnary mapping a hash code into knowed objects.
            restore_id: If True, the id will be restored using the given data.
                If False, a new id will be generated.

        """
        raise NotImplementedError()

    def complete_with_default(self, data: OrderedDict) -> None:
        """Add default data in place when fields are missing."""
        for key in self.MANDATORY_FIELDS:
            if key not in data:
                raise ValueError(f"{key} of the socket is missing")

        for key, val in self.DEFAULT_DATA.items():
            if key not in data:
                data[key] = val
